#!/usr/bin/env python

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Entry point for pipelines run within Docker images.
"""
import logging
import os
import sys

import argparse
import disdat.common
import disdatluigi.common
import disdat.fs
import disdat.api
import disdatluigi.api
from disdat import log

import boto3
from botocore.exceptions import ClientError

_HELP = """ Run a Disdat pipeline. This script wraps up several of the
steps required to run a pipeline, including: creating a working context, 
running a pipeline class to generate an output bundle, and pushing an 
output bundle to a Disdat remote.
"""

_logger = logging.getLogger(__name__)


def _context_and_remote(context_name, remote=None):
    """Create a new Disdat context and bind remote if not None.

    Check environment for 'LOCAL_EXECUTION', which should exist and be True if we are running
    a container in an existing .disdat environment (e.g., on someone's laptop).

    If so, do not take any actions that would change the state of the users CLI.  That is, do not
    switch contexts.

    Args:
        context_name (str): A fully-qualified context name. remote-context/local-context or local-context
        remote (str): S3 remote name.
    """

    retval = disdat.api.context(context_name)

    if retval == 1: # branch exists
        _logger.warning("Entrypoint found existing local context {} ".format(context_name))
        _logger.warning("Entrypoint not switching and ignoring directive to change to remote context {}".format(remote))
    elif retval == 0: # just made a new branch
        if remote is not None:
            _logger.info("Entrypoint made a new context {}, attaching remote {}".format(context_name, remote))
            _remote(context_name, remote)
    else:
        _logger.error("Entrypoint got non standard retval {} from api.context({}) command.".format(retval, context_name))
        return False

    if disdatluigi.common.LOCAL_EXECUTION not in os.environ:
        disdat.api.switch(context_name)
    else:
        _logger.info("Container running locally (not in a cloud provider, aka AWS).  Not switching contexts")

    return True


def _remote(context_arg, remote_url):
    """ Add remote to our context.

    Args:
        context_arg:  <remote context>/<local context> or <local context> to use in this container
        remote_url: The remote to add to this local context

    Returns:
        None
    """
    _logger.debug("Adding remote at URL {} for branch '{}'".format(remote_url, context_arg))

    contexts = context_arg.split('/')

    if len(contexts) > 1:
        remote_context = contexts[0]
        local_context = contexts[1]
    else:
        local_context = contexts[0]
        remote_context = local_context

    if remote_url is None:
        _logger.error("Got an invalid URL {}".format(remote_url))
        return False

    try:
        disdat.api.remote(local_context, remote_context, remote_url)
    except Exception:
        return False
    return True


def retrieve_secret(secret_name):
    """ Placeholder for ability to retrieve secrets needed by image

    Returns:

    """

    raise NotImplementedError

    # Modify these to get them from the current environment
    endpoint_url = "https://secretsmanager.us-west-2.amazonaws.com"
    region_name = "us-west-2"

    session = boto3.session.Session()
    client = session.client(
        service_name='secretsmanager',
        region_name=region_name,
        endpoint_url=endpoint_url
    )

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            print("The requested secret " + secret_name + " was not found")
        elif e.response['Error']['Code'] == 'InvalidRequestException':
            print(("The request was invalid due to:", e))
        elif e.response['Error']['Code'] == 'InvalidParameterException':
            print(("The request had invalid params:", e))
    else:
        # Decrypted secret using the associated KMS CMK
        # Depending on whether the secret was a string or binary, one of these fields will be populated
        if 'SecretString' in get_secret_value_response:
            secret = get_secret_value_response['SecretString']
        else:
            binary_secret_data = get_secret_value_response['SecretBinary']

        print ("Found the secret string as ")
        print(secret)


def add_argument_help_string(help_string, default=None):
    if default is None:
        return '{}'.format(help_string)
    else:
        return "{} (default '{}')".format(help_string, default)


def _commit_and_push(b):
    """ commit and push bundle b if not transient """
    if disdat.common.BUNDLE_TAG_TRANSIENT not in b.tags:
        b.commit()
        b.push()


def run_disdat_container(args):
    """ Execute Disdat inside of container

    Args:
        args: input arguments

    Returns:
        None

    """
    print("Entrypoint running with args: {}".format(args))

    if args.remote is not None:
        response = boto3.client('sts').get_caller_identity()
        _logger.info("boto3 caller identity {}".format(response))
        incremental_pull = True   # running with a remote
    else:
        incremental_pull = False  # running without a remote

    print ("Entrypoint running with incremental_pull=={}".format(incremental_pull))

    # Check to make sure that we have initialized the Disdat environment
    if not os.path.exists(os.path.join(os.environ['HOME'], '.config', 'disdat')):
        _logger.warning("Disdat environment possibly uninitialized?")

    # Create context, add remote, and switch to it
    if not _context_and_remote(args.branch, args.remote):
        _logger.error("Failed to branch to \'{}\' and optionally bind  to \'{}\'".format(args.branch,
                                                                                         args.remote))
        sys.exit(os.EX_IOERR)

    # Pull the remote branch into the local branch or download individual items
    try:
        if not args.no_pull:
            disdat.api.pull(args.branch, localize=not incremental_pull)
    except Exception as e:
        _logger.error("Failed to pull and localize all bundles from context {} due to {}".format(args.branch, e))
        sys.exit(os.EX_IOERR)

    # If specified, decode the ordinary 'key:value' strings into a dictionary of tags.
    input_tags = disdat.common.parse_args_tags(args.input_tag)
    output_tags = disdat.common.parse_args_tags(args.output_tag)

    # Convert string of pipeline args into dictionary for api.apply
    deser_user_params = disdatluigi.common.parse_params(args.pipe_cls, args.pipeline_args)

    # If the user wants final and intermediate, then inc push.
    if not args.no_push and not args.no_push_intermediates:
        incremental_push = True
    else:
        incremental_push = False

    try:
        result = disdatluigi.api.apply(args.branch,
                                  args.pipe_cls,
                                  output_bundle=args.output_bundle,
                                  input_tags=input_tags,
                                  output_tags=output_tags,
                                  params=deser_user_params,
                                  output_bundle_uuid=args.output_bundle_uuid,
                                  force=args.force,
                                  force_all=args.force_all,
                                  workers=args.workers,
                                  incremental_push=incremental_push,
                                  incremental_pull=incremental_pull)

        if not incremental_push:
            if not args.no_push:
                if not args.no_push_intermediates:
                    to_push = disdat.api.search(args.branch, is_committed=False, find_intermediates=True)
                    for b in to_push:
                        _commit_and_push(b)
                b = disdat.api.get(None, uuid=args.output_bundle_uuid)
                if b is not None:
                    _logger.info("Pipeline ran.  Committing and pushing output bundle UUID {}.".format(args.output_bundle_uuid))
                    _commit_and_push(b)
                else:
                    _logger.info("Pipeline ran but did not finish final task.")
            else:
                _logger.info("Pipeline ran but user specified not to push any bundles to remote context.")
        else:
            _logger.info("Pipeline ran using incremental pushing.")

    except RuntimeError as re:
        _logger.error('Failed to run pipeline: RuntimeError {}'.format(re))
        sys.exit(os.EX_IOERR)

    except disdatluigi.common.ApplyError as ae:
        _logger.error('Failed to run pipeline: ApplyException {}'.format(ae))
        sys.exit(os.EX_IOERR)

    if args.dump_output:
        print(disdat.api.cat(args.branch, args.output_bundle))

    sys.exit(os.EX_OK)


def argparse_and_run(input_args):

    # To simplify configuring and building pipeline images, we can keep
    # various default parameter values in the Docker image makefile,
    # and pass them on as Docker ENV variables.   At the moment, we set
    # the default params below to handle most cases.  This is an example
    # of how you might do this in the future if needed.
    # some_default = os.environ[ENVVAR] if ENVVAR in os.environ else None

    parser = argparse.ArgumentParser(
        description=_HELP,
    )

    parser.add_argument(
        '--dump-output',
        help='Dump the output to standard output',
        action='store_true',
    )
    parser.add_argument(
        '--debug-level',
        default=logging.INFO,
        help='The debug logging level (default {})'.format(logging.getLevelName(logging.WARNING))
    )

    disdat_parser = parser.add_argument_group('remote repository arguments')
    disdat_parser.add_argument(
        '--remote',
        type=str,
        default=None,
        help='The s3 bucket from/to which to pull/push data',
    )
    disdat_parser.add_argument(
        '--no-pull',
        action='store_true',
        help='Do not pull (synchronize) remote repository with local repo - may cause entire pipeline to re-run.',
    )
    disdat_parser.add_argument(
        '--no-push',
        action='store_true',
        help='Do not push output bundles (including intermediates) to the remote repository (default is to push)',
    )
    disdat_parser.add_argument(
        '--no-push-intermediates',
        action='store_true',
        help='Do not push the intermediate bundles to the remote repository (default is to push)',
    )

    pipeline_parser = parser.add_argument_group('pipe arguments')

    pipeline_parser.add_argument(
        '--branch',
        type=str,
        required=True,
        help='The fully-qualified Disdat branch to use when running',
    )

    pipeline_parser.add_argument(
        '--workers',
        type=int,
        default=2,
        help="The number of Luigi workers to spawn.  Default is 2."
    )

    pipeline_parser.add_argument(
        '-it', '--input-tag',
        nargs=1, type=str, action='append',
        help="Input bundle tags: '-it authoritative:True -it version:0.7.1'")

    pipeline_parser.add_argument(
        '-ot', '--output-tag',
        nargs=1, type=str, action='append',
        help="Output bundle tags: '-ot authoritative:True -ot version:0.7.1'")

    pipeline_parser.add_argument(
        '--output-bundle-uuid',
        default=None,
        type=str,
        help='UUID for the output bundle (default is for apply to generate a UUID)',
    )

    pipeline_parser.add_argument(
        '-o',
        '--output-bundle',
        type=str,
        default='-',
        help="Name output bundle: '-o my.output.bundle'.  Default name is '<TaskName>_<param_hash>'"
    )

    pipeline_parser.add_argument(
        '--force',
        action='store_true',
        help='Force recomputation of the last task.',
    )

    pipeline_parser.add_argument(
        '--force-all',
        action='store_true',
        help='Force recomputation of all upstream tasks.',
    )

    pipeline_parser.add_argument(
        'pipe_cls',
        default=None,
        type=disdat.common.load_class,
        help=add_argument_help_string("Name of the pipeline class to run, e.g., 'package.module.ClassName'"),
    )

    pipeline_parser.add_argument(
        "pipeline_args",
        nargs=argparse.REMAINDER,
        type=str,
        help="Optional set of parameters for this pipe '--parameter value'"
    )

    args = parser.parse_args(input_args)

    log_level = logging.INFO

    log.enable(level=log_level)  # TODO: Add configurable verbosity

    run_disdat_container(args)


def main():
    input_args = sys.argv[1:]
    argparse_and_run(input_args)


if __name__ == '__main__':
    main()





#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

import boto3
import moto
import docker
import pytest

from tests.conftest import TEST_CONTEXT
from tests.functional.common_tasks import COMMON_DEFAULT_ARGS
import disdat.api as disdat_api
import disdatluigi.api as api


TEST_NAME    = 'test_bundle'
TEST_BUCKET  = 'test-bucket'
TEST_BUCKET_URL = "s3://{}".format(TEST_BUCKET)
PWD          = os.path.dirname(__file__)
SETUP_DIR    = os.path.join(PWD,'..')
PIPELINE_CLS = 'functional.common_tasks.A'


@pytest.fixture(scope="module")
def build_container_setup_only():
    """ Create a docker image locally.  At the moment we are only
    testing whether the basic python setup.py builds correctly.

    TODO: Need to test using the configure directory with
        a.) Using a MANIFEST file
        b.) Installing custom python packages
        c.) Installing rpms
        d.) Installing R packages
    """

    retval = api.dockerize(SETUP_DIR)
    id = api.dockerize_get_id(SETUP_DIR)
    yield id
    docker_client = docker.from_env()
    docker_client.images.remove(id, force=True)


def test_run_local_container(run_test, build_container_setup_only):
    """ Run the local container.
    Test if it runs, test if it re-runs all steps, test if it re-runs last step.
    """

    retval = api.run(SETUP_DIR,
                     TEST_CONTEXT,
                     PIPELINE_CLS
                     )

    b_b = disdat_api.get(TEST_CONTEXT, 'B')
    assert b_b is not None

    b_a = disdat_api.get(TEST_CONTEXT, 'A')
    assert b_a is not None
    assert b_a.data == sum(COMMON_DEFAULT_ARGS)

    # Re-run with force all

    retval = api.run(SETUP_DIR,
                     TEST_CONTEXT,
                     PIPELINE_CLS,
                     {'int_array': [1, 2, 3]},
                     force_all=True
                     )

    b_b_f = disdat_api.get(TEST_CONTEXT, 'B')
    assert b_b_f is not None
    assert b_b.uuid != b_b_f.uuid

    b_a_f = disdat_api.get(TEST_CONTEXT, 'A')
    assert b_a_f is not None
    assert b_a.uuid != b_a_f.uuid
    assert b_a_f.data == sum([1, 2, 3])

    # Re-run with force last one

    retval = api.run(SETUP_DIR,
                     TEST_CONTEXT,
                     PIPELINE_CLS,
                     {'int_array': [1, 2, 3]},
                     force=True
                     )

    b_b_f2 = disdat_api.get(TEST_CONTEXT, 'B')
    assert b_b_f2 is not None
    assert b_b_f.uuid == b_b_f2.uuid

    b_a_f2 = disdat_api.get(TEST_CONTEXT, 'A')
    assert b_a_f2 is not None
    assert b_a_f.uuid != b_a_f2.uuid


#@moto.mock_aws
def manual_test_run_aws_batch(run_test, build_container_setup_only):
    """ Incomplete test.   The container code itself needs to have
    its S3 access mocked out.  Here we are testing manually
    """

    # Setup moto s3 resources
    #s3_resource = boto3.resource('s3')
    #s3_resource.create_bucket(Bucket=TEST_BUCKET)

    # Add a remote.   Pull and Push!
    manual_s3_url = 's3://'
    disdat_api.remote(TEST_CONTEXT, TEST_CONTEXT, manual_s3_url)

    retval = api.run(SETUP_DIR,
                     TEST_CONTEXT,
                     PIPELINE_CLS,
                     remote_context=TEST_CONTEXT,
                     remote_s3_url=manual_s3_url,
                     pull=True,
                     push=True
                     )

    # Blow away everything and pull
    disdat_api.rm(TEST_CONTEXT, bundle_name='.*', rm_all=True)
    disdat_api.pull(TEST_CONTEXT)
    b = disdat_api.get(TEST_CONTEXT, 'A')
    assert b.data == sum(COMMON_DEFAULT_ARGS)


if __name__ == '__main__':
    pytest.main([__file__])



from setuptools import find_packages, setup

setup(
    name='disdat-test-pipelines',
    version=0.1,
    packages=find_packages(exclude=['config']),
    include_package_data=True,
    install_requires=[
        'pandas'
    ],
)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABCMeta, abstractmethod
import os

import luigi
from luigi.contrib.s3 import S3Target
from six.moves import urllib

from disdat.fs import DisdatFS

from disdatluigi import logger as _logger


MISSING_EXT_DEP_UUID = 'UnresolvedExternalDep'
YIELD_PIPETASK_ARG_NAME = "YieldArgName"

class PipeBase(object):
    __metaclass__ = ABCMeta

    @property
    def pfs(self):
        return DisdatFS()

    @abstractmethod
    def output_bundle(self):
        """
        Given this pipe, return the set of bundles created by this pipe.
        Mirrors Luigi task.outputs()

        Returns:
            (processing_name, uuid)
        """
        pass

    @abstractmethod
    def input_bundles(self):
        """
        Given this pipe, return the set of bundles created by the input pipes.
        Mirrors Luigi task.inputs()

        :param pipe_task:  A PipeTask or a DriverTask (both implement PipeBase)
        Returns:
            [(processing_name, uuid), ... ]
        """
        pass

    @abstractmethod
    def processing_id(self):
        """
        Given a pipe instance, return a unique string based on the class name and
        the parameters.

        Bundle Tag:   Used to fill in bundle.processing_name
        """
        pass

    @abstractmethod
    def human_id(self):
        """
        This is a "less unique" id than the unique id.  It is supposed to be the "human readable" name of the stage
        this pipe occupies in the pipesline.

        Bundle Tag:   Used to fill in bundle.bundle_name
        """
        pass

    @staticmethod
    def _interpret_scheme(full_path):
        scheme = urllib.parse.urlparse(full_path).scheme

        if scheme == '' or scheme == 'file':
            ''' LOCAL FILE '''
            return luigi.LocalTarget(full_path)
        elif scheme == 's3':
            ''' S3  FILE '''
            return S3Target(full_path)

        assert False

    @staticmethod
    def filename_to_luigi_targets(output_dir, output_value):
        """
        Create Luigi file objects from a file name, dictionary of file names, or list of file names.

        Return the same object type as output_value, but with Luigi.Targets instead.

        Args:
            output_dir (str): Managed output path.
            output_value (str, dict, list): A basename, dictionary of basenames, or list of basenames.

        Return:
            (`luigi.LocalTarget`, `luigi.contrib.s3.S3Target`): Singleton, list, or dictionary of Luigi Target objects.
        """

        if isinstance(output_value, list) or isinstance(output_value, tuple):
            luigi_outputs = []
            for i in output_value:
                full_path = os.path.join(output_dir, i)
                luigi_outputs.append(PipeBase._interpret_scheme(full_path))
            if len(luigi_outputs) == 1:
                luigi_outputs = luigi_outputs[0]
        elif isinstance(output_value, dict):
            luigi_outputs = {}
            for k, v in output_value.items():
                full_path = os.path.join(output_dir, v)
                luigi_outputs[k] = PipeBase._interpret_scheme(full_path)
        else:
            full_path = os.path.join(output_dir, output_value)
            luigi_outputs = PipeBase._interpret_scheme(full_path)

        return luigi_outputs



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os
import argparse
import multiprocessing

import luigi.task_register
from luigi import build
from luigi.execution_summary import LuigiStatusCode, _partition_tasks

import disdat.common as common  # config, especially logging, before luigi ever loads
import disdat.fs as fs
from disdatluigi.common import apply_handle_result, parse_params
from disdatluigi import logger as _logger


def apply(output_bundle, pipe_params, pipe_cls, input_tags, output_tags, force, force_all,
          output_bundle_uuid=None, central_scheduler=False, workers=1, data_context=None,
          incremental_push=False, incremental_pull=False):
    """
    Given an input bundle, run the pipesline on the bundle.
    Note, we first make a copy of all tasks that are parameterized identically to the tasks we will run.
    This is so we can figure out what we will need to re-run.
    This is why we make a single uuid for the output bundle of apply (for the driver).

    Args:
        output_bundle: The new bundle to be created
        pipe_params (dict):   mapping of parameter name to Luigi Task parameter value
        pipe_cls (type[disdat.pipe.PipeTask]):      reference to the task class
        force:         force recomputation of the last user task.
        force_all:     force recomputation of dependencies
        input_tags (dict):  Tags used to find the input bundle
        output_tags (dict):  Tags that need to be placed on the output bundle
        force_all (bool): whether to re-run this pipe
        output_bundle_uuid (str):  Optionally specify exactly the UUID of the output bundle IFF we actually need to produce it
        central_scheduler: Use a centralized Luigi scheduler (default False, i.e., --local-scheduler is used)
        workers: The number of luigi workers to use for this workflow (default 1)
        data_context: Actual context object or None and read current context.
        incremental_push (bool): Whether this job should push tasks as they complete to the remote (if configured)
        incremental_pull (bool): Whether this job should localize bundles as needed from the remote (if configured)

    Returns:
        bool: True if there were no failed tasks and no failed schedulings (missing external dependencies)
    """

    _logger.debug("pipe_cls {}".format(pipe_cls))
    _logger.debug("pipe params: {}".format(pipe_params))
    _logger.debug("force: {}".format(force))
    _logger.debug("force_all: {}".format(force_all))
    _logger.debug("input tags: {}".format(input_tags))
    _logger.debug("output tags: {}".format(output_tags))
    _logger.debug("sys.path {}".format(sys.path))
    _logger.debug("central_scheduler {}".format(central_scheduler))
    _logger.debug("workers {}".format(workers))
    _logger.debug("incremental_push {}".format(incremental_push))
    _logger.debug("incremental_pull {}".format(incremental_pull))

    if incremental_push:
        _logger.warning("incremental_push {}".format(incremental_push))

    if incremental_pull:
        _logger.warning("incremental_pull {}".format(incremental_pull))

    pfs = fs.DisdatFS()

    if data_context is None:
        if not pfs.in_context():
            _logger.warning('Not in a data context')
            return None
        data_context = pfs.curr_context

    # Increment the reference count for this process
    apply.reference_count += 1

    # If we are using Fork, we cannot have internal apply's with workers > 1
    # otherwise luigi loops forever with "There are no more tasks to run" and "<some task> is currently run by worker"
    # This happens with Vanilla luigi in fork mode.   In <=P37, MP fork for OS X is the default
    # in >=P38, Spawn is the default.
    if apply.reference_count > 1:
        if multiprocessing.get_start_method() == 'fork':
            workers = 1

    def cleanup_cached_state():
        """
        After running, decrement our reference count (which tells how many simultaneous apply methods are
        running nested in this process.  Once the last one completes, blow away the luigi instance cache and git hash.
        Needed if we're run twice (from scratch) in the same process.  Otherwise, on the next run, we could find
        the same class instances, with the old cached_output_bundle fields set.
        """
        apply.reference_count -= 1
        if not apply.reference_count:
            fs.DisdatFS().clear_pipe_version()
            luigi.task_register.Register.clear_instance_cache()

    # Only pass data_context name, not reference to the pipe class
    # data contexts may have open sql connections and other state
    # that is not ForingPickler safe.
    data_context_name = data_context.get_local_name()

    # Re-execute logic -- make copy of task DAG
    # Creates a cache of {pipe:path_cache_entry} in the pipesFS object.
    # This "task_path_cache" is used throughout execution to find output bundles.
    dag = create_users_task(pipe_cls, pipe_params, output_bundle,
                            output_bundle_uuid, force_all, output_tags,
                            data_context_name, incremental_push, incremental_pull)

    # Get version information for pipeline
    pipeline_path = os.path.dirname(sys.modules[dag.__module__].__file__)
    fs.DisdatFS().get_pipe_version(pipeline_path)

    # If the user just wants to re-run this task, use mark_force
    if force:
        dag.mark_force()

    # Will be LuigiRunResult
    status = build([dag], local_scheduler=not central_scheduler, workers=workers, detailed_summary=True)
    success = False
    if status.status == LuigiStatusCode.SUCCESS:
        success = True
    task_sets = _partition_tasks(status.worker)
    did_work = len(task_sets['completed']) > 0

    cleanup_cached_state()

    return {'success': success, 'did_work': did_work}


# Add a reference count to apply, so we can determine when to clean up the path_cache
apply.reference_count = 0


def create_users_task(pipe_cls,
                      pipe_params,
                      root_bundle_name,
                      forced_output_bundle_uuid,
                      force,
                      output_tags,
                      data_context_name,
                      incremental_push,
                      incremental_pull):
    """
    Create the users task

    Every apply or run is logically a single execution and produces a set of outputs.  Each
    task produces a single bundle represented as a hyperframe internally.

    Args:
        pipe_cls (disdat.Pipe): The user's Pipe class
        pipe_params: parameters for this pipeline
        root_bundle_name (str): user set output bundle name for last task
        forced_output_bundle_uuid (str): user set output bundle uuid for last task, else None
        force (bool): force re-run of entire pipeline
        output_tags (dict):  str,str tag dict
        data_context_name (str): name in which pipe will run
        incremental_push (bool): push bundle when pipe finishes
        incremental_pull (bool): pull non-localized bundles before execution

    Returns:
        `disdat.Pipe`
    """

    # Force root task to take an explicit bundle name?
    if root_bundle_name == '-':
        root_bundle_name = None

    task_params = {'is_root_task':  True,
                   'root_output_bundle_name': root_bundle_name,
                   'forced_output_bundle_uuid': forced_output_bundle_uuid,
                   'force': force,
                   'output_tags': output_tags,
                   'data_context_name': data_context_name,
                   'incremental_push': incremental_push,
                   'incremental_pull': incremental_pull
                   }

    # Get user pipeline parameters for this Pipe / Luigi Task
    if pipe_params:
        task_params.update(pipe_params)

    # Instantiate and return the class directly with the parameters
    # Instance caching is taken care of automatically by luigi
    return pipe_cls(**task_params)


def cli_apply(args):
    """
    Parse and prepare strings from argparse arguments into suitable Python objects
    to call the api's version of apply.  Note, args.pipe_cls is already a cls object.
    Most of the work here is to deser each input parameter value according to its
    Luigi definition.

    Parameters:
        disdat_config:
        args:

    Returns:
        None
    """

    if not fs.DisdatFS().in_context():
        print("Apply unavailable -- Disdat not in a valid context.")
        return

    # Create a dictionary of str->str arguments to str->python objects deser'd by Luigi Parameters
    deser_user_params = parse_params(args.pipe_cls, args.params)

    input_tags = common.parse_args_tags(args.input_tag)

    output_tags = common.parse_args_tags(args.output_tag)

    # NOTE: sysexit=False is required for us to pass a data_context object through luigi tasks.
    # Else we build up arguments as strings to run_with_retcodes().  And it crashes because the data_context is
    # not a string.
    result = apply(args.output_bundle, deser_user_params, args.pipe_cls, input_tags, output_tags,
                   args.force, args.force_all,
                   central_scheduler=args.central_scheduler,
                   workers=args.workers,
                   incremental_push=args.incremental_push,
                   incremental_pull=args.incremental_pull)

    # If we didn't successfully run any task, sys.exit with non-zero code
    apply_handle_result(result)


def add_arg_parser(subparsers):
    """Initialize a command line set of subparsers with file system commands.

    Args:
        subparsers: A collection of subparsers as defined by `argsparse`.
    """

    apply_p = subparsers.add_parser('apply',
                                    description="Apply a transform to an input bundle to produce an output bundle.")
    apply_p.add_argument('-cs', '--central-scheduler', action='store_true', default=False,
                         help="Use a central Luigi scheduler (defaults to local scheduler)")
    apply_p.add_argument('-w', '--workers', type=int, default=1, help="Number of Luigi workers on this node")
    apply_p.add_argument('-it', '--input-tag', nargs=1, type=str, action='append',
                         help="Input bundle tags: '-it authoritative:True -it version:0.7.1'")
    apply_p.add_argument('-ot', '--output-tag', nargs=1, type=str, action='append',
                         help="Output bundle tags: '-ot authoritative:True -ot version:0.7.1'")
    apply_p.add_argument('-o', '--output-bundle', type=str, default='-',
                         help="Name output bundle: '-o my.output.bundle'.  Default name is '<TaskName>_<param_hash>'")
    apply_p.add_argument('-f', '--force', action='store_true', help="Force re-computation of only this task.")
    apply_p.add_argument('--force-all', action='store_true', help="Force re-computation of ALL upstream tasks.")
    apply_p.add_argument('--incremental-push', action='store_true',
                         help="Commit and push each task's bundle as it is produced to the remote.")
    apply_p.add_argument('--incremental-pull', action='store_true',
                         help="Localize bundles as they are needed by downstream tasks from the remote.")
    apply_p.add_argument('pipe_cls', type=common.load_class, help="User-defined transform, e.g., 'module.PipeClass'")
    apply_p.add_argument('params', type=str, nargs=argparse.REMAINDER,
                         help="Optional set of parameters for this pipe '--parameter value'")
    apply_p.set_defaults(func=lambda args: cli_apply(args))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
A PipeTask is a wrapper around a Luigi Task

It simplifies the Luigi Task interface in a couple of ways.
1.) requires():
  luigi.Task.requires() becomes disdat_luigi.PipeTask.pipe_requires
  In Disdat-Luigi, you call self.add_dependency() and self.add_external_dependency()
  to add upstream tasks.
2.) run():
  luigi.Task.run() becomes disdat_luigi.PipeTask.pipe_run()
  In Disdat-Luigi, pipe_run() receives a parameter list of the objects returned by
  the upstream tasks.   Those parameter names are set via the self.add_dependency calls in pipe_requires().
  Thus the return value matters in pipe_run().
  See https://disdat.gitbook.io/disdat-documentation/basic-concepts/bundles/untitled
3.) Users no longer need to specify an output() function in their task.
"""

import os
import sys
import time
import hashlib
from types import GeneratorType

import luigi
from luigi import worker
from disdat.common import BUNDLE_TAG_TRANSIENT, BUNDLE_TAG_PUSH_META
from disdat.hyperframe import HyperFrameRecord
from disdat.fs import DisdatFS
from disdat.utility.bundle_helpers import different_code_versions
import disdat.api as api

from disdatluigi.pipe_base import PipeBase, YIELD_PIPETASK_ARG_NAME, MISSING_EXT_DEP_UUID
from disdatluigi.common import ExtDepError
from disdatluigi import logger as _logger


class PipeTask(luigi.Task, PipeBase):
    """
    user_arg_name:
    calling_task:
    driver_output_bundle:
    force:
    output_tags:
    incremental_push:
    incremental_pull:
    """
    user_arg_name = luigi.Parameter(default=None, significant=False)  # what the outputs are referred to by downstreams
    is_root_task   = luigi.BoolParameter(default=None, significant=False)

    root_output_bundle_name = luigi.Parameter(default=None, significant=False)
    forced_output_bundle_uuid = luigi.Parameter(default=None, significant=False)

    force = luigi.BoolParameter(default=False, significant=False)
    output_tags = luigi.DictParameter(default={}, significant=False)

    # Each pipeline executes wrt a data context.
    data_context_name = luigi.Parameter(default=None, significant=False)

    # Each pipeline can be configured to commit and push intermediate values to the remote
    incremental_push = luigi.BoolParameter(default=False, significant=False)

    # Each pipeline can be configured to pull intermediate values on demand from the remote
    incremental_pull = luigi.BoolParameter(default=False, significant=False)

    def __init__(self, *args, **kwargs):
        """
        This has the same signature as luigi.Task.
        Go through this class and get the set of params we define

        Args:
            *args:
            **kwargs:
        """

        super(PipeTask, self).__init__(*args, **kwargs)

        # Luigi may copy the task by first serializing an existing task using task.to_str_params()
        # This forces string parameters set to None to 'None'
        if self.root_output_bundle_name == 'None':
            self.root_output_bundle_name = None
        if self.forced_output_bundle_uuid == 'None':
            self.forced_output_bundle_uuid = None

        self._cached_processing_id = None
        self._cached_output_bundle = None  # refers to either a new bundle or existing bundle: b.uuid and b.local_dir
        self._is_new_output = None

        # Instance variables to track various user wishes
        self.user_set_human_name = None  # self.set_bundle_name()
        self.user_tags = {}              # self.add_tags()
        self.add_deps = {}               # self.add(_external)_dependency()
        self.yield_deps = {}             # self.yield_dependency()
        self._input_tags = {}            # self.get_tags() of upstream tasks
        self._input_bundle_uuids = {}    # self.get_bundle_uuid() of upstream tasks
        self._mark_force = False         # self.mark_force()

    def input_bundles(self):
        """
        Given this pipe, return the set of bundles that this task used as input.
        Return a list of tuples that contain (processing_name, uuid, arg_name)

        NOTE: Calls task.deps which calls task._requires which calls task.requires()

        Args:
            self (disdatluigi.PipeTask):  The pipe task in question

        Returns:
            (dict(str:`disdat.api.Bundle`)):  {arg_name: bundle, ...}
        """

        input_bundles = {}
        for task in self.deps():
            b = task.output_bundle
            assert b is not None
            input_bundles[task.user_arg_name] = b
        return input_bundles

    @property
    def output_bundle(self):
        """ Resolve this task to an existing bundle or get a new bundle.

        If this is a new bundle (a bundle does not yet exist for this processing name)
        then we will create an output directory and cache a pointer to this bundle in this object.
        Otherwise cache the closed bundle with the result.

        However, in the case of yielded tasks, Luigi re-creates the task by serializing the parameters and
        deserializing them during task creation, often creating a new object.  Now we have another task
        that represents this same data (and processing name), but when the original asks for its output bundle it
        doesn't have the *same* uuid that the new task has.   Yikes!

        So we need a remote db to store the processingname -> bundle uuid binding.  It's the data context (and
        the sqlite db).  But we only push to the db on bundle write (it's either closed / on disk / in db, or nothing).
        So storing an additional bundle state would require some changes to the db.

        In lieu of that, we adopt the following slightly more expensive policy:
        * if cache empty: cache returned bundle from resolve.
        * if cached and closed: return bundle
        * if cached and open:  call resolve.  [Did an identical task write this logical bundle (by processing name)?]
        *     if new_bundle is closed:  set cached to closed bundle
        *     elif new_bundle is open: abondon bundle.
        *      if closed: return cached closed bundle.  There should be no situation where resolve returns a different bundle.

        It is possible that two tasks do the same work and write to two different bundles with the same processing name.
        This won't be incorrect, just inefficient.   Future fix would be to synchronize on the DB vis-a-vis above.
        """
        if self._cached_output_bundle is None:
            self._cached_output_bundle = self._resolve_bundle(self.pfs.get_context(self.data_context_name))
            if not self._cached_output_bundle.closed:
                self._cached_output_bundle.open(force_uuid=self.forced_output_bundle_uuid)
        else:
            if not self._cached_output_bundle.closed:
                possible_output = self._resolve_bundle(self.pfs.get_context(self.data_context_name))
                # <------- RACE ------->   between not closed and check, i.e., not thread safe.
                if possible_output.closed:
                    if possible_output.uuid == self._cached_output_bundle.uuid:
                        # Note, we cannot "clean up" the bundle if we GC one. See Bundle.abandon()
                        del self._cached_output_bundle
                    else:
                        self._cached_output_bundle.abandon()
                    self._cached_output_bundle = possible_output

        return self._cached_output_bundle

    @property
    def pipe_output(self):
        """ Return the data in the bundle presented as the Python type with which it was stored """
        return self.output_bundle.data

    @property
    def is_new_output(self):
        if self._is_new_output is None:
            self._is_new_output = not self.output_bundle.closed
        return self._is_new_output

    def processing_id(self):
        """
        Given a pipe instance, return the "processing_name" -- a unique string based on the class name and
        the parameters.  This re-uses Luigi code for getting the unique string.

        processing_id = task_class_name + hash(task params +  [ dep0_task_name + hash(dep0_params),
                        dep1_task_name + hash(dep1_params),...])

        In typical Luigi workflows, a task's parameters uniquely determine the requires function outputs.

        Thus there is no need to include the data inputs in the uniqueness of the processing name.

        But, some facilities on top of Disdat might build dynamic classes.   In these cases, upper layer code
        might define a pipeline in python like:
        ```
        @punch.task
        def adder(a,b):
            return a+b

        def pipeline(X):
            u = adder(a=1, b=2)
            v = adder(a=u, b=X+1)
            w = adder(a=v, b=X+2)
        ```

        In this case, there is a new task for each `adder`.  Each`adder`'s requirements are defined in a scope outside
         of the dynamic classes requires method.

        NOTE: The PipeTask has a 'driver_output_bundle'.  This the name of the pipeline output bundle given by the user.
        Because this is a Luigi parameter, it is included in the Luigi self.task_id string and hash.  So we do not
        have to append this separately.

        """

        if self._cached_processing_id is not None:
            return self._cached_processing_id

        deps = self.requires()

        assert(isinstance(deps, dict))

        input_task_processing_ids = [(deps[k]).processing_id() for k in sorted(deps)]

        as_one_str = '.'.join(input_task_processing_ids)

        param_hash = hashlib.md5(as_one_str.encode('utf-8')).hexdigest()

        processing_id = self.task_id + '_' + param_hash[:luigi.task.TASK_ID_TRUNCATE_HASH]

        self._cached_processing_id = processing_id

        return processing_id

    def human_id(self):
        """
        This is the "human readable" name;  a "less unique" id than the unique id.

        The pipeline_id is well-defined for the output task -- it is the output bundle name.   For intermediate outputs
        the pipeline_id defaults to the pipe_id().   Else, it may be set by the task author.

        Note: Should we provide an identify for which version of this pipe is running at which stage in the pipeline?
        Short answer, no.   Imagine if we name with the pipeline bundle output name, branch index, and level index.  In
        this case if anyone re-uses this output, the human_name for the bundle won't be meaningful.   For the pipeline
        owner, it may also not be helpful.  The system may also place different outputs at different times under those
        indices.  Too complicated.

        Returns:
            (str)

        """

        if self.root_output_bundle_name is not None:
            return self.root_output_bundle_name
        elif self.user_set_human_name is not None:
            return self.user_set_human_name
        else:
            default_human_name = type(self).__name__
            return "{}".format(default_human_name)

    def get_hframe_uuid(self):
        """ Return the unique ID for this tasks current output hyperframe

        Returns:
            hframe_uuid (str): The unique identifier for this task's hyperframe
        """
        return self.output_bundle.uuid

    def upstream_hframes(self):
        """ Convert upstream tasks to hyperframes, return list of hyperframes

        Returns:
            (:list:`hyperframe.HyperFrameRecord`): list of upstream hyperframes
        """

        tasks = self.deps()
        hfrs = []
        for t in tasks:
            hfid = t.get_hframe_uuid()
            hfrs.append(self.pfs.get_hframe_by_uuid(hfid, data_context=self.pfs.get_context(self.data_context_name)))

        return hfrs

    def requires(self):
        """
        Return Tasks on which this task depends.

        Build them intelligently, however.
        1.) The input_df so far stays the same for all upstream pipes.
        2.) However, when we resolve the location of the outputs, we need to do so correctly.

        Args:

        Returns:
            (dict): arg_name: task_class
        """

        kwargs = self.prepare_pipe_kwargs()

        self.add_deps.clear()
        self.pipe_requires(**kwargs)
        rslt = self.add_deps

        if len(self.add_deps) == 0:
            return {}

        tasks = {}

        for user_arg_name, cls_and_params in rslt.items():
            pipe_class, params = cls_and_params[0], cls_and_params[1]
            assert isinstance(pipe_class, luigi.task_register.Register)
            self._update_dependency_pipe_params(params, user_arg_name)
            tasks[user_arg_name] = pipe_class(**params)

        return tasks

    def _update_dependency_pipe_params(self, params, user_arg_name):
        """
        Given a parameter dictionary, update with PipeTask parameters
        Never called on the user's root task.  It's important to delete
        root_output_bundle_name and forced_output_bundle_uuid b/c if someone
        yields a PipeTask, Luigi doesn't take the instance, they make a new one
        and the parameter serializer replaces None with 'None' for string params.

        Args:
            params (dict): a dictionary of str: arg for a new PipeTask class
            user_arg_name (str): The user's name of the resulting task output

        Returns:
            (dict)
        """
        # we propagate the same inputs and the same output dir for every upstream task!
        params.update({
            'user_arg_name': user_arg_name,
            'is_root_task': False,
            'force': self.force,
            'output_tags': dict({}),  # do not pass output_tags up beyond root task
            'data_context_name': self.data_context_name,  # all operations wrt this context
            'incremental_push': self.incremental_push,  # propagate the choice to push incremental data.
            'incremental_pull': self.incremental_pull  # propagate the choice to incrementally pull data.
        })

        return params

    def output(self):
        """
        This is the *only* output function for all pipes.  It declares the creation of the
        one HyperFrameRecord pb and that's it.  Remember, has to be idempotent.

        Note: By checking self.output_bundle we are implicitly also looking up our output bundle.

        Return:
            (list:str):

        """
        if self.output_bundle is None:
            assert(self.uuid == MISSING_EXT_DEP_UUID)  # only reason we should be here.
            output_path = DisdatFS().get_context(self.data_context_name).get_object_dir()
            output_uuid = "this_file_should_not_exist"  # look for a file that should not exist
        else:
            output_path = self.output_bundle.local_dir
            output_uuid = self.output_bundle.uuid

        return luigi.LocalTarget(os.path.join(output_path, HyperFrameRecord.make_filename(output_uuid)))

    def run(self):
        """
        Call users run function.
        1.) prepare the arguments
        2.) run and gather user result
        3.) interpret and wrap in a HyperFrame

        Returns:
            None
        """
        kwargs = self.prepare_pipe_kwargs(for_run=True)
        bundle = self.output_bundle
        assert(bundle is not None)

        """ NOTE: If a user changes a task param in run(), and that param parameterizes a dependency in requires(), 
        then running requires() post run() will give different tasks.  To be safe we record the inputs before run() 
        """
        add_dep_bundle_inputs = self.input_bundles()

        try:
            start = time.time()  # P3 datetime.now().timestamp()
            user_rtn_val = self.pipe_run(**kwargs)
            if isinstance(user_rtn_val, GeneratorType):
                try:
                    while True:  # while the user's task is yielding
                        task_list = next(user_rtn_val)
                        if not isinstance(task_list, list): task_list = [task_list]
                        # have we been here before?  If so, don't yield.
                        if all(task.complete() for task in task_list):
                            pass
                        else:
                            yield task_list  # yield the task or list of tasks
                except StopIteration as si:
                    user_rtn_val = si.value # we always have a return value, it is in si.value
            stop = time.time()  # P3 datetime.now().timestamp()
        except Exception as error:
            """ If user's pipe fails for any reason, remove bundle dir and raise """
            try:
                _logger.error("User pipe_run encountered exception: {}".format(error))
                bundle.abandon()
            except OSError as ose:
                _logger.error("User pipe_run encountered error, and error on remove bundle: {}".format(ose))
            raise

        try:
            # Add any output tags to the user tag dict
            if self.output_tags:
                self.user_tags.update(self.output_tags)

            # If this is the root_task, identify it as so in the tag dict
            if self.is_root_task:
                self.user_tags.update({'root_task': 'True'})

            """ if we have a pce, we have a new bundle that we need to add info to and close """
            bundle.add_data(user_rtn_val)

            bundle.add_timing(start, stop)

            bundle.add_dependencies(add_dep_bundle_inputs.values(), add_dep_bundle_inputs.keys())

            yield_output_bundles = [t.output_bundle for t in self.yield_deps.values()]
            yield_output_names = self.yield_deps.keys()
            bundle.add_dependencies(yield_output_bundles, yield_output_names)

            bundle.name = self.human_id()

            bundle.processing_name = self.processing_id()

            bundle.add_params(self._get_subcls_params())

            bundle.add_tags(self.user_tags)

            bundle.add_code_ref('{}.{}'.format(self.__class__.__module__, self.__class__.__name__))

            pipeline_path = os.path.dirname(sys.modules[self.__class__.__module__].__file__)
            cv = DisdatFS.get_pipe_version(pipeline_path)
            bundle.add_git_info(cv.url, cv.hash, cv.branch)

            bundle.close()  # Write out the bundle

            """ Incrementally push the completed bundle """
            if self.incremental_push and (BUNDLE_TAG_TRANSIENT not in bundle.tags):
                self.pfs.commit(None, None, uuid=bundle.uuid, data_context=self.pfs.get_context(self.data_context_name))
                self.pfs.push(uuid=bundle.uuid, data_context=self.pfs.get_context(self.data_context_name))

        except Exception as error:
            """ If we fail for any reason, remove bundle dir and raise """
            bundle.abandon()
            raise

        return None

    def _get_subcls_params(self):
        """ Given the child class, extract user defined Luigi parameters

        This function uses vars(cls) and filters by Luigi Parameter
        types.  Luigi get_params() gives us all parameters in the full class hierarchy.
        It would give us the parameters in this class as well.  And then we'd have to do set difference.
        See luigi.Task.get_params()

        Args:
            self: The instance of the subclass.  To get the normalized values for the Luigi Parameters
        Returns:
            dict: {<name>:'string value',...}
        """
        cls = self.__class__
        params = {}
        for param in vars(cls):
            attribute = getattr(cls, param)
            if isinstance(attribute, luigi.Parameter):
                params["{}".format(param)] = attribute.serialize(getattr(self, param))
        return params

    @classmethod
    def _put_subcls_params(cls, ser_params):
        """ Given the child class, create the Luigi parameter dictionary

        Assume that ser_params dictionary keys are the attribute names in the Disdat task class.

        Args:
            self: The instance of the subclass.  To get the normalized values for the Luigi Parameters
            ser_params (dict): Dictionary <str>:<str>
        Returns:
            deser_params (dict): {<name>: Luigi.Parameter,...}
        """
        deser_params = {}
        for param, ser_value in ser_params.items():
            try:
                attribute = getattr(cls, param)
                assert isinstance(attribute, luigi.Parameter)
                deser_params[param] = attribute.parse(ser_value)
            except Exception as e:
                _logger.warning("Bundle parameter ({}:{}) can't be deserialized by class({}): {}".format(param,
                                                                                                         ser_value,
                                                                                                         cls.__name__,
                                                                                                         e))
                raise e
        return deser_params

    def prepare_pipe_kwargs(self, for_run=False):
        """ Each upstream task produces a bundle.  Prepare that bundle as input
        to the user's pipe_run function.

        Args:
            for_run (bool): prepare args for run -- at that point all upstream tasks have completed.

        Returns:
            (dict): A dictionary with the arguments.

        """
        kwargs = dict()

        # Place upstream task outputs into the kwargs.  Thus the user does not call
        # self.inputs().  If they did, they would get a list of output targets for the bundle
        if for_run:

            # Reset the stored tags, in case this instance is run multiple times.
            self._input_tags = {}
            self._input_bundle_uuids = {}

            upstream_tasks = [(t.user_arg_name, t.output_bundle) for t in self.deps()]

            for user_arg_name, b in [u for u in upstream_tasks if u[1] is not None]:
                assert b.is_presentable

                # Download data that is not local (the linked files not present). Default when running in container.
                if self.incremental_pull:
                    b.pull(localize=True)

                if user_arg_name in kwargs:
                    _logger.warning('Task human name {} reused when naming task dependencies'.format(user_arg_name))

                self._input_tags[user_arg_name] = b.tags
                self._input_bundle_uuids[user_arg_name] = b.uuid
                kwargs[user_arg_name] = b.data

        return kwargs

    """
    Pipes Interface -- A pipe implements these calls
    """

    def pipe_requires(self, **kwargs):
        """
        This is the place to put your pipeline dependencies.  Place
        the upstream pipes in an array and a dict for their params

        Args:
            **kwargs:

        Returns:

        """
        return None

    def pipe_run(self, **kwargs):
        """
        There is only one default argument "input_df" in kwargs.
        The other keys in kwargs will be identical to your Luigi parameters specified in this class.

        The input_df has the data context identifiers, e.g., sampleName, sessionId, subjectId
        The input_df has the data in either jsonData or fileData.
        A sharded task will receive a subset of all possible inputs.

        Args:
            **kwargs:

        Returns:

        """
        raise NotImplementedError()

    def yield_dependency(self, pipe_class, params):
        """
        Disdat Pipe API Function

        Use this function to dynamically yield dependencies in the pipe_run function.  Access these dependency
        results by retaining a reference to the PipeTask and calling PipeTask.pipe_output.
        ``
        def pipe_run(self, some_arg=None):
            pipe = self.yield_dependency(IsFish, params=None)
            yield pipe
            if pipe.pipe_output is True:
              print("It's a fish.")
        ``

        Args:
            pipe_class (object):  Class name of upstream task if looking for external bundle by processing_id.
            params (dict):  Dictionary of parameters if looking for external bundle by processing_id.

        Returns:
            `disdatluigi.PipeTask`
        """
        if not isinstance(params, dict):
            error = "yield_dependency: params argument must be a dictionary"
            raise Exception(error)

        assert isinstance(pipe_class, luigi.task_register.Register)

        # Note: we instantiate a new class each time to get the Luigi Task.task_id.  we don't need the
        # PipeTask.processing_name because we only need tasks that are unique to this parent task (the one yielding).
        # But we can't just use yielded task order to name them if the task yields in different orders.
        self._update_dependency_pipe_params(params, YIELD_PIPETASK_ARG_NAME)
        to_yield = pipe_class(**params)
        id = to_yield.task_id

        if id not in self.yield_deps:
            self.yield_deps[id] = to_yield
        else:
            if self.yield_deps[id] is not to_yield: del to_yield

        return self.yield_deps[id]

    def add_dependency(self, param_name, pipe_class, params):
        """
        Disdat Pipe API Function

        Add a task and its parameters to our requirements

        Args:
            param_name (str): The parameter name this bundle assumes when passed to Pipe.run
            pipe_class (object):  Class name of upstream task if looking for external bundle by processing_id.
            params (dict):  Dictionary of parameters if looking for external bundle by processing_id.

        Returns:
            None

        """
        if not isinstance(params, dict):
            error = "add_dependency third argument must be a dictionary of parameters"
            raise Exception(error)

        assert (param_name not in self.add_deps)
        self.add_deps[param_name] = (pipe_class, params)

        return

    def add_external_dependency(self, param_name, pipe_class, params, human_name=None, uuid=None):
        """
        Disdat Pipe API Function

        Add an external task and its parameters to our requirements.   What this means is that
        there is no run function and, in that case, Luigi will ignore the results of task.deps() (which calls
        flatten(self.requires())).  And what that means is that this requirement can only be satisfied
        by the bundle actually existing.

        Create ersatz ExternalDepTask parameterized by uuid and processing_name
        Note: it is possible to use class/params when searching by class, params, but this makes all external
        dependencies look the same in the code.  Win.

        NOTE: if you add an external dependency by name, it is possible that someone adds a bundle during
        execution and that your requires function is no longer deterministic.   You must add caching to your
        requires function to handle this scenario.

        Example with class variable bundle_uuid:
        ``
        if self.bundle_uuid is None:
            bundle = self.add_external_dependency('_', MyTaskClass, {}, human_name='some_result')
            self.bundle_uuid = bundle.uuid
        else:
            bundle = self.add_external_dependency('_', MyTaskClass, {}, uuid=self.bundle_uuid)
        ``

        TODO: Consider pushing caching into this layer.

        Args:
            param_name (str): The parameter name this bundle assumes when passed to Pipe.run
            pipe_class (object):  Class name of upstream task if looking for external bundle by processing_id.
            params (dict):  Dictionary of parameters if looking for external bundle by processing_id.
            human_name (str): Resolve dependency by human_name, return the latest bundle with that humman_name.  Trumps task_class and params.
            uuid (str): Resolve dependency by explicit UUID, trumps task_class, params and human_name.

        Returns:
            `api.Bundle` or None

        """
        import disdat.api as api

        if pipe_class is not None and not isinstance(params, dict):
            error = "add_external_dependency requires parameter dictionary"
            raise Exception(error)

        assert (param_name not in self.add_deps)

        try:
            if uuid is not None:
                hfr = self.pfs.get_hframe_by_uuid(uuid, data_context=self.pfs.get_context(self.data_context_name))
            elif human_name is not None:
                hfr = self.pfs.get_latest_hframe(human_name, data_context=self.pfs.get_context(self.data_context_name))
            else:
                # we propagate the same inputs and the same output dir for every upstream task!
                params.update({
                    'user_arg_name': param_name,
                    'data_context_name': self.data_context_name
                })
                p = pipe_class(**params)
                hfr = self.pfs.get_hframe_by_proc(p.processing_id(), data_context=self.pfs.get_context(self.data_context_name))

            if hfr is None:
                error_str = "Disdat can't resolve external bundle from class[{}] params[{}] name[{}] uuid[{}]".format(pipe_class,
                                                                                                                      params,
                                                                                                                      human_name,
                                                                                                                      uuid)
                raise ExtDepError(error_str)

            bundle = api.Bundle(self.data_context_name).fill_from_hfr(hfr)

        except ExtDepError as error:  # Swallow and allow Luigi to determine task is not available.
            _logger.error(error_str)
            bundle = None

        except Exception as error:
            _logger.error(error)
            bundle = None

        finally:
            if bundle is None:
                self.add_deps[param_name] = (luigi.task.externalize(ExternalDepTask), {'uuid': MISSING_EXT_DEP_UUID,
                                                                                       'processing_name': 'None'})
            else:
                # When a task requires an external dep, this can be called multiple times.  And then
                # we return to the pipe.run which creates the class.  Note that calling task.deps() will cause
                # the requires() to be called.  But calling deps() doesn't mean that task.output() will be called
                # And that means is that tasks that have been required might not have their task.cached_output_bundle
                # set by calling resolve_bundle.  Now, in this case, because we are calling luigi.task.externalize,
                # we are actually creating *copies* of the class object and so luigi object caching isn't going to work.
                # This means that resolve_bundle must be called when using the cached_output_bundle field.
                self.add_deps[param_name] = (luigi.task.externalize(ExternalDepTask), {'uuid': bundle.uuid,
                                                                                       'processing_name': bundle.processing_name})

        return bundle

    def create_output_file(self, filename):
        """
        Disdat Pipe API Function

        Pass in the name of your file, and get back a Luigi target object to which you can write.

        Args:
            filename (str, dict, list): A basename, dictionary of basenames, or list of basenames.

        Returns:
            (`luigi.LocalTarget`): Singleton, list, or dictionary of Luigi Target objects.
        """
        output_dir = self.output_bundle.local_dir

        return self.filename_to_luigi_targets(output_dir, filename)

    def create_remote_output_file(self, filename):
        """
        Disdat Pipe API Function

        Pass in the name of your file, and get back an object to which you can write on S3.

        NOTE: Managed S3 paths are created only if a) remote is set (otherwise where would we put them?)
        and b) incremental_push flag is True  (if we don't push bundle metadata, then the locations may be lost).

        Args:
            filename (str, dict, list): A basename, dictionary of basenames, or list of basenames.

        Returns:
            (`luigi.contrib.s3.S3Target`): Singleton, list, or dictionary of Luigi Target objects.

        """
        output_dir = self.get_remote_output_dir()
        return self.filename_to_luigi_targets(output_dir, filename)

    def create_output_dir(self, dirname):
        """
        Disdat Pipe API Function

        Given basename directory name, return a fully qualified path whose prefix is the
        local output directory for this bundle in the current context.  This call creates the
        output directory as well.

        Args:
            dirname (str): The name of the output directory, i.e., "models"

        Returns:
            output_dir (str):  Fully qualified path of a directory whose prefix is the bundle's local output directory.

        """
        prefix_dir = self.get_output_dir()
        fqp = os.path.join(prefix_dir, dirname)
        try:
            os.makedirs(fqp)
        except IOError as why:
            _logger.error("Creating directory in bundle directory failed:".format(why))

        return fqp

    def create_remote_output_dir(self, dirname):
        """
        Disdat Pipe API Function

        Given basename directory name, return a fully qualified path whose prefix is the
        remote output directory for this bundle in the current context.

        NOTE: The current context must have a remote attached.

        Args:
            dirname (str): The name of the output directory, i.e., "models"

        Returns:
            output_dir (str):  Fully qualified path of a directory whose prefix is the bundle's remote output directory.

        """
        prefix_dir = self.get_remote_output_dir()
        fqp = os.path.join(prefix_dir, dirname)
        return fqp

    def get_output_dir(self):
        """
        Disdat Pipe API Function

        Retrieve the output directory for this task's bundle.  You may place
        files directly into this directory.

        Returns:
            output_dir (str):  The bundle's output directory

        """
        return self.output_bundle.local_dir

    def get_remote_output_dir(self):
        """
        Disdat Pipe API Function

        Retrieve the output directory for this task's bundle.  You may place
        files directly into this directory.

        Returns:
            output_dir (str):  The bundle's output directory on S3

        """
        uuid = self.output_bundle.uuid

        data_context = self.pfs.get_context(self.data_context_name)
        if data_context.remote_ctxt_url and self.incremental_push:
            output_dir = os.path.join(data_context.get_remote_object_dir(), uuid)
        else:
            raise Exception('Managed S3 path creation needs a) remote context and b) incremental push to be set')
        return output_dir

    def set_bundle_name(self, human_name):
        """
        Disdat Pipe API Function

        Set the human name for this bundle.  If not called, then intermediate outputs
        will have human names identical to their process names.

        Args:
            human_name (str):  The human name of this pipe's output bundle.

        Returns:
            None

        """
        self.user_set_human_name = human_name

    def add_tags(self, tags):
        """
        Disdat Pipe API Function

        Adds tags to bundle.

        Args:
            tags (dict (str, str)): key value pairs (string, string)

        Returns:
            None
        """
        assert (isinstance(tags, dict))
        self.user_tags.update(tags)

    def get_tags(self, user_arg_name):
        """
        Disdat Pipe API Function

        Retrieve the tag dictionary from an upstream task.

        Args:
            user_arg_name (str): keyword arg name of input bundle data for which to return tags

        Returns:
            tags (dict (str, str)): key value pairs (string, string)
        """
        assert user_arg_name in self._input_tags
        return self._input_tags[user_arg_name]

    def get_bundle_uuid(self, user_arg_name):
        """
        Disdat Pipe API Function

        Retrieve the UUID from an upstream task.

        Args:
            user_arg_name (str): keyword arg name of input bundle data for which to return tags

        Returns:
            uuid (str)
        """
        assert user_arg_name in self._input_bundle_uuids
        return self._input_bundle_uuids[user_arg_name]

    def mark_force(self):
        """
        Disdat Pipe API Function

        Mark pipe to force recompution of this particular task.   This means that Disdat/Luigi will
        always re-run this particular pipe / task.

        We mark the pipe with a particular flag so that apply.resolve_bundle()

        Returns:
            None
        """
        self._mark_force = True

    def mark_transient(self, push_meta=True):
        """
        Disdat Pipe API Function

        Mark output bundle as transient.   This means that during execution Disdat will not
        write (push) this bundle back to the remote.  That only happens in two cases:
        1.) Started the pipeline with incremental_push=True
        2.) Running the pipeline in a container with no_push or no_push_intermediates False

        We mark the bundle with a tag.   Incremental push investigates the tag before pushing.
        And the entrypoint investigates the tag if we are not pushing incrementally.
        Otherwise, normal push commands from the CLI or api will work, i.e., manual pushes continue to work.

        Args:
            push_meta (bool): Push the meta-data but not the data. Else, push nothing.

        Returns:
            None
        """
        if push_meta:
            self.add_tags({BUNDLE_TAG_TRANSIENT: 'True', BUNDLE_TAG_PUSH_META: 'True'})
        else:
            self.add_tags({BUNDLE_TAG_TRANSIENT: 'True'})

    def _resolve_bundle(self, data_context):
        """
        Instead of resolving before we run, this resolve can be issued
        from within the pipe.output() function.

        Note: Only returns None with an unfound external dependency

        Note: If returning a new bundle, we do *not* open it in this function. Bundle.open()
        creates the output directory.  Only do that if we know we're going to actually use the
        bundle in pipe.output().

        Args:
            self: the pipe to investigate
            data_context: the data context object from which we should resolve bundles.

        Returns:
            Bundle: bundle to use. If open, new bundle, if closed, re-using
        """
        verbose = False

        if verbose:
            print("resolve_bundle: looking up bundle {}".format(self.processing_id()))

        if (self._mark_force or self.force) and not isinstance(self, ExternalDepTask):
            # Forcing recomputation through a manual annotation or --force directive, unless external
            _logger.debug("resolve_bundle: pipe.mark_force forcing a new output bundle.")
            if verbose: print("resolve_bundle: pipe.mark_force forcing a new output bundle.\n")
            return api.Bundle(data_context)

        if isinstance(self, ExternalDepTask):
            # NOTE: Even if add_external_dependency() fails to find the bundle we still succeed here.
            # Thus it can look like we reuse a bundle, when in fact we don't.  We error either
            # within the user's requires, add_external_dependency(), or when Luigi can't find the task (current approach)
            assert worker._is_external(self)
            b = api.get(data_context.get_local_name(), None, uuid=str(self.uuid))
            if b is not None:
                if verbose: print("resolve_bundle: found ExternalDepTask re-using bundle with UUID[{}].\n".format(self.uuid))
            else:
                if verbose: print("resolve_bundle: No ExternalDepTask found with UUID[{}].\n".format(self.uuid))
            return b

        bndls = api.search(data_context.get_local_name(), processing_name=self.processing_id())

        if bndls is None or len(bndls) <= 0:
            if verbose: print("resolve_bundle: No bundle with proc_name {}, getting new output bundle.\n".format(self.processing_id()))
            return api.Bundle(data_context)

        bndl = bndls[0]  # our best guess is the most recent bundle with the same processing_id()

        # 2.) Bundle exists - lineage object tells us input bundles.
        lng = bndl.get_lineage()
        if lng is None:
            if verbose: print("resolve_bundle: No lineage present, getting new output bundle.\n")
            return api.Bundle(data_context)

        # 3.) Lineage record exists -- if new code, re-run
        pipeline_path = os.path.dirname(sys.modules[self.__module__].__file__)
        current_version = DisdatFS().get_pipe_version(pipeline_path)

        if different_code_versions(current_version, lng):
            if verbose: print("resolve_bundle: New code version, getting new output bundle.\n")
            return api.Bundle(data_context)

        # 3.5.) Have we changed the output human bundle name?  If so, re-run task.
        # Note: we need to go through all the bundle versions with that processing_id.
        # because, at the moment, we make new bundles when we change name.  When in some sense
        # it's just a tag set that should include other names and the data should be the same.

        current_human_name = self.human_id()
        found = False
        for bndl in bndls:
            if current_human_name == bndl.get_human_name():
                found = True
                break
        if not found:
            if verbose: print("resolve_bundle: New human name {} (prior {}), getting new output bundle.\n".format(
                current_human_name, bndl.get_human_name()))
            return api.Bundle(data_context)

        # 4.) Check the inputs -- assumes we have processed upstream tasks already
        for task in self.deps():
            """ Are we re-running an upstream input?
            Look through its *current* list of possible upstream tasks, not the ones it had
            on its prior run.   If the UUID has changed relative to lineage, then we need to re-run.
            
            In general, the only reason we should re-run an upstream is b/c of a code change.  And that change
            did not change the tasks parameters.  So it looks the same, but it is actually different.  OR someone 
            deletes and re-runs (maybe sql query result changes though parameters are the same).
            
            But if an output exists and we want to ignore code version and ignore data changes then
            while we do this, we should re-use our bundle independent of whether an upstream needs to re-run 
            or whether one of our inputs is out of date. 
            
            So one option is to ignore upstreams that need to be re-run.  Re-use blindly.  Like Luigi.  
            
            Another option is that anytime we don't have an input bundle, we attempt to read it not just
            locally, but remotely as well.               
            """
            LUIGI_RERUN = False

            if LUIGI_RERUN:
                # Ignore whether upstreams had to be re-run b/c they didn't have bundles.
                # Ignore whether this has to be re-run because existing inputs are newer
                continue

            if task.output_bundle is None:
                # this can happen with bundles created by other pipelines.
                # still surface the warning, but no longer raise exception
                _logger.info(
                    "Resolve bundles: this pipe's dep {} has no input bundle. Likely an externally produced bundle".format(
                        task.processing_id()))
            else:
                if task.is_new_output:
                    if verbose: print("Resolve_bundle: upstream task is being re-run, so rerun with new output bundle.\n")
                    return api.Bundle(data_context)

                # Upstream Task
                # Resolve to a bundle, UUID and a processing name
                # If it is an ordinary task in a workflow, we resolve via the processing name
                if worker._is_external(task) and isinstance(task, ExternalDepTask):
                    upstream_dep_uuid = task.uuid
                    upstream_dep_processing_name = task.processing_name
                else:
                    found = api.search(data_context.get_local_name(), processing_name=task.processing_id())
                    assert len(found) > 0
                    local_bundle = found[0]  # the most recent with this processing_name
                    upstream_dep_uuid = local_bundle.pb.uuid
                    upstream_dep_processing_name = local_bundle.pb.processing_name
                    assert(upstream_dep_processing_name == task.processing_id())

                """ Now we need to check if we should re-run this task because an upstream input exists and has been updated        
                Go through each of the inputs used for this current task.  
                POLICY
                1.) if the date is more recent, it is "new" data.
                2.) if it is older, we should require force (but currently do not and re-run).
                XXX TODO: Add date to the depends_on pb data structure to enforce 2 XXX
                """
                for tup in lng.pb.depends_on:
                    if tup.hframe_proc_name == upstream_dep_processing_name and tup.hframe_uuid != upstream_dep_uuid:
                        if verbose: print("Resolve_bundle: prior input bundle {} uuid {} has new uuid {}\n".format(
                            task.processing_id(),
                            tup.hframe_uuid,
                            upstream_dep_uuid))
                        return api.Bundle(data_context)

        # 5.) Woot!  Reuse the found bundle.
        if verbose: print("resolve_bundle: reusing bundle\n")
        return bndl


class ExternalDepTask(PipeTask):
    """ This task is only here as a shell.
    If the user specifies an external dependency, we look it up in add_external_dependency.
    We look it up, b/c we want to hand them the bundle in requires.
    If they look up the bundle via UUID or human name, there is no reason for them to
    pass in the class.  Especially for human name, where they cannot know it.
    And, if it exists, there is no reason to look it up again in apply.resolve_bundle().
    Thus we create an ExternalDepTask() parameterized by the UUID and apply.resolve_bundle()
    The default output() function will create this tasks processing_id() which will be a hash
    of this task's params, which will include a unique UUID.   And so should be unique.
    """
    uuid = luigi.Parameter(default=None)
    processing_name = luigi.Parameter(default=None)

    def input_bundles(self):
        """ External bundles output are in lineage.
        Note: this is only called in apply for now.  And this task can never
        be called by apply directly.
        """
        assert False, "An ExternalDepTask should never be run directly."



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import pytest
import disdat.api as api

TEST_CONTEXT = '___test_context___'
    
"""
NOTE: When testing we try to use Moto to mock-out the s3 calls.  Pyarrow sidesteps boto by using its own S3FileSystem.  This is a problem
because that c++ code is no longer accessing s3 URLs through the botocore Python library.  So you get errors about INVALID CREDENTIALS. 
The primary solution for testing is to simply use fastparquet instead of pyarrow.  The other alternative is to use the moto_server or, as the 
comment below suggests, use fsspec's s3fs.  At this time (7-2024) we simply use fastparquet for testing.

From: https://issues.apache.org/jira/browse/ARROW-16437?page=com.atlassian.jira.plugin.system.issuetabpanels%3Acomment-tabpanel&focusedCommentId=17530795#comment-17530795

    I don't think you can use the "mock_s3" method of moto directly with our S3 filesystem integration (as Antoine said, we don't use boto). 
    A possible way to do this is to use fsspec's s3fs, and then pass this filesystem object (and pyarrow will wrap that in a PyFileSystem), 
    but that adds another layer of indirection which you don't need here (and which can be another source of failures).
    Another option is to use the "moto_server" feature of moto (http://docs.getmoto.org/en/latest/docs/getting_started.html#stand-alone-server-mode), 
    which gives you an endpoint url, that can be used to construct a pyarrow S3FileSystem that interacts with the moto server.
    This is basically the approach that we use ourselves (but with MinIO instead of moto), and eg also dask switched from mock_s3 to 
    moto_server in their tests (see eg https://github.com/dask/dask/blob/4d6a5f08c45be56302f696ca4ef6038a1cd1e734/dask/bytes/tests/test_s3.py#L84)

"""

@pytest.fixture(autouse=True, scope='module')
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    _aws_credentials()

def _aws_credentials():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"

@pytest.fixture(autouse=True, scope='function')
def run_test():
    # Remove test context before running test
    if TEST_CONTEXT in api.ls_contexts():
        api.delete_context(context_name=TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)
    yield
    api.delete_context(context_name=TEST_CONTEXT)



#
# Copyright 2015, 2016, 2017 Human Longevity, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Utilities for accessing AWS using boto3

@author: twong / kyocum
@copyright: Human Longevity, Inc. 2017
@license: Apache 2.0
"""

import base64
import pkg_resources
from getpass import getuser
import six

from botocore.exceptions import ClientError
import boto3 as b3

from disdatluigi import logger as _logger
from disdat.utility.aws_s3 import profile_get_region

def batch_get_job_definition_name(pipeline_image_name):
    """Get the most recent active AWS Batch job definition for a dockerized
    pipeline.

    Note: The Python getpass docs do not specify the exception thrown when getting the user name fails.

    """

    try:
        return '{}-{}-job-definition'.format(getuser(), pipeline_image_name)
    except Exception as e:
        return '{}-{}-job-definition'.format('DEFAULT', pipeline_image_name)


def batch_get_latest_job_definition(job_definition_name):
    """Get the most recent active revision number for a AWS Batch job
    definition

    Args:
        job_definition_name: The name of the job definition
        remote_pipeline_image_name:
        vcpus:
        memory:

    Return:
        The latest job definition dictionary or `None` if the job definition does not exist
    """
    region = profile_get_region()
    client = b3.client('batch', region_name=region)
    response = client.describe_job_definitions(jobDefinitionName=job_definition_name, status='ACTIVE')
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        raise RuntimeError(
            'Failed to get job definition revisions for {}: HTTP Status {}'.format(job_definition_name, response['ResponseMetadata']['HTTPStatusCode'])
        )
    job_definitions = response['jobDefinitions']
    revision = 0
    job_def = None
    for j in job_definitions:
        if j['jobDefinitionName'] != job_definition_name:
            continue
        if j['revision'] > revision:
            revision = j['revision']
            job_def = j

    return job_def


def batch_extract_job_definition_fqn(job_def):
    revision = job_def['revision']
    name = job_def['jobDefinitionName']
    return '{}:{}'.format(name, revision)


def batch_get_job_definition(job_definition_name):
    """Get the most recent active revision number for a AWS Batch job
    definition

    Args:
        job_definition_name: The name of the job definition

    Return:
        The fully-qualified job definition name with revision number, or
            `None` if the job definition does not exist
    """
    job_def = batch_get_latest_job_definition(job_definition_name)

    if job_def is None:
        return None
    else:
        return batch_extract_job_definition_fqn(job_def)


def batch_register_job_definition(job_definition_name, remote_pipeline_image_name,
                                  vcpus=1, memory=2000, job_role_arn=None):
    """Register a new AWS Batch job definition.

    Args:
        job_definition_name: The name of the job definition
        remote_pipeline_image_name: The ECR Docker image to load to run jobs
            using this definition
        vcpus: The number of vCPUs to use to run jobs using this definition
        memory: The amount of memory in MiB to use to run jobs using this
            definition
        job_role_arn (str): Can be None
    """

    container_properties = {
        'image': remote_pipeline_image_name,
        'vcpus': vcpus,
        'memory': memory,
    }

    if job_role_arn is not None:
        container_properties['jobRoleArn'] = job_role_arn

    region = profile_get_region()
    client = b3.client('batch', region_name=region)
    response = client.register_job_definition(
        jobDefinitionName=job_definition_name,
        type='container',
        containerProperties=container_properties
    )
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        raise RuntimeError('Failed to create job definition {}: HTTP Status {}'.format(job_definition_name, response['ResponseMetadata']['HTTPStatusCode']))

    return response


def ecr_create_fq_respository_name(repository_name, policy_resource_package=None, policy_resource_name=None):
    ecr_client = b3.client('ecr', region_name=profile_get_region())
    # Create or fetch the repository in AWS (to store the image)
    try:
        response = ecr_client.create_repository(
            repositoryName=repository_name
        )
        repository_metadata = response['repository']
        # Set the policy on the repository
        if policy_resource_package is not None and policy_resource_name is not None:
            policy = pkg_resources.resource_string(policy_resource_package.__name__, policy_resource_name)
            _ = ecr_client.set_repository_policy(
                registryId=repository_metadata['registryId'],
                repositoryName=repository_name,
                policyText=policy,
                force=True
            )
    except ClientError as e:
        if e.response['Error']['Code'] == 'RepositoryAlreadyExistsException':
            response = ecr_client.describe_repositories(
                repositoryNames=[repository_name]
            )
            repository_metadata = response['repositories'][0]
        elif e.response['Error']['Code'] == 'AccessDeniedException':
            _logger.warning("Error [AccessDeniedException] when creating repo {}, trying to continue...".format(repository_name))
        else:
            raise e
    return repository_metadata['repositoryUri']


def ecr_get_fq_repository_name(repository_name):
    return ecr_create_fq_respository_name(repository_name)


def ecr_get_auth_config():
    ecr_client = b3.client('ecr', region_name=profile_get_region())
    # Authorize docker to push to ECR
    response = ecr_client.get_authorization_token()
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        raise RuntimeError('Failed to get AWS ECR authorization token: HTTP Status {}'.format(response['ResponseMetadata']['HTTPStatusCode']))
    token = response['authorizationData'][0]['authorizationToken']

    token_bytes = six.b(token)

    token_decoded_bytes = base64.b64decode(token_bytes)

    token_decoded_str = token_decoded_bytes.decode('utf8')

    username, password = token_decoded_str.split(':')

    return {'username': username, 'password': password}



#!/usr/bin/env python

import argparse
import importlib as imp
import mock
import os
import setuptools


def find_packages(setup_py):
    packages = []
    # All this horrid hackery to recover the install_requires parameter from
    # a setup() call in setup.py.
    #
    # https://stackoverflow.com/questions/24236266/how-to-extract-dependencies-information-from-a-setup-py
    try:
        # Patch setuptools to intercept the setup() call
        with mock.patch.object(setuptools, 'setup') as setup_mock:
            # Get an open file handle and a description of the
            # setup file.
            setup_file, setup_filename, setup_description = imp.find_module('setup', [os.path.dirname(setup_py)])
            # Load setup.py as the module setup. We have to
            # intercept calls to find_packages as well since
            # find_packages will run a 'find'-like operation from
            # the current working directory - which is Bad if the
            # CWD is the root directory...
            with mock.patch.object(setuptools, 'find_packages'):
                imp.load_module('setup', setup_file, setup_filename, setup_description)
        # Grab the call args to setup
        _, setup_kwargs = setup_mock.call_args
        # ...and recover the install_requires parameter. Fun, eh?
        # Don't forget to remove trailing version specifiers that
        # lack version numbers.
        packages = ['{}'.format(p.rstrip('<>=')) for p in setup_kwargs['install_requires']]
    finally:
        # As warned in the docs, we have to close the setup file
        # ourselves.
        if setup_file is not None:
            setup_file.close()
    return packages


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Select Python packages from setup.py'
    )
    parser.add_argument(
        'setup_py',
        type=str,
        help='The setup.py file',
    )
    args = parser.parse_args()

    if not os.path.exists(args.setup_py):
        raise RuntimeError('Failed to find file {}'.format(args.setup_py))

    print ('\n'.join(find_packages(args.setup_py)))


#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test Incremental Push

Use API to create a bundle with some files
push to remote context

author: Sayantan Satpati
"""
import boto3
import moto
from moto import mock_aws
import pandas as pd
import pytest
import os

from tests.conftest import _aws_credentials

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT

TEST_REMOTE = '__test_remote_context__'
TEST_BUCKET = 'test-bucket'
TEST_BUCKET_OTHER = 'test-bucket-other'
TEST_BUCKET_URL = "s3://{}".format(TEST_BUCKET)

# ================================================== #
# ================ Disdat Tasks ==================== #
# ================================================== #


class NonManagedLocal(PipeTask):
    def pipe_requires(self):
        self.set_bundle_name('b1')

    def pipe_run(self):
        # Task output was created on some local path
        d = {'col1': [1, 2], 'col2': [3, 4]}
        df = pd.DataFrame(data=d)
        local_file = '/tmp/test.parquet'
        df.to_parquet(local_file)
        return {'file': [local_file]}


class NonManagedS3(PipeTask):
    def pipe_requires(self):
        self.set_bundle_name('b2')

    def pipe_run(self):
        # Task output was created on some S3 path
        d = {'col1': [1, 2], 'col2': [3, 4]}
        df = pd.DataFrame(data=d)
        s3_file = 's3://{}/test.parquet'.format(TEST_BUCKET_OTHER)
        df.to_parquet(s3_file)

        return {'file': [s3_file]}


class ManagedLocal(PipeTask):
    def pipe_requires(self):
        self.set_bundle_name('b3')

    def pipe_run(self):
        target = self.create_output_file('test.parquet')

        # Write dataframe to S3 Managed Path
        d = {'col1': [1, 2], 'col2': [3, 4]}
        df = pd.DataFrame(data=d)
        with target.temporary_path() as temp_output_path:
            df.to_parquet(temp_output_path)

        return {'file': [target.path]}


class ManagedS3(PipeTask):
    def pipe_requires(self):
        self.set_bundle_name('b4')

    def pipe_run(self):
        target = self.create_remote_output_file('test.parquet')

        # Write dataframe to S3 Managed Path
        d = {'col1': [1, 2], 'col2': [3, 4]}
        df = pd.DataFrame(data=d)
        with target.temporary_path() as temp_output_path:
            df.to_parquet(temp_output_path)

        return {'file': [target.path]}


# ================================================== #
# ================== Tests ========================= #
# ================================================== #


def test_managed_local():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    dlapi.apply(TEST_CONTEXT, ManagedLocal)
    assert len(api.search(TEST_CONTEXT)) == 1, 'Only one bundle should be present'
    print(api.cat(TEST_CONTEXT, 'b3'))

    assert os.path.exists(api.search(TEST_CONTEXT, human_name='b3')[0].data['file'][0]), \
        'Local file should be present in bundle'


def test_non_managed_local():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    dlapi.apply(TEST_CONTEXT, NonManagedLocal)
    assert len(api.search(TEST_CONTEXT)) == 1, 'Only one bundle should be present'
    print(api.cat(TEST_CONTEXT, 'b1'))

    assert os.path.exists(api.search(TEST_CONTEXT, human_name='b1')[0].data['file'][0]), \
        'Local file should be present in bundle'


@moto.mock_aws
def test_remote_push_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Bind remote context
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    # Apply
    dlapi.apply(TEST_CONTEXT, ManagedS3, incremental_push=True)

    assert not os.path.exists(api.search(TEST_CONTEXT, human_name='b4')[0].data['file'][0]), \
        'Managed S3 file should not be copied to local'

    # Get objects from remote
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    keys = [o['Key'] for o in objects['Contents']]
    keys = [key.split('/')[-1] for key in keys]

    # Make sure files exist in S3
    for output_file in ['test.parquet']:
        assert output_file in keys, 'Pipeline should have pushed file'


@moto.mock_aws
def test_remote_push_non_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)
    s3_resource.create_bucket(Bucket=TEST_BUCKET_OTHER)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'
    objects = s3_client.list_objects(Bucket=TEST_BUCKET_OTHER)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Bind remote context
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    # Apply
    dlapi.apply(TEST_CONTEXT, NonManagedS3, incremental_push=True)
    print(api.cat(TEST_CONTEXT, 'b2'))

    # Local context should not contain file if a remote exists.
    b = api.search(TEST_CONTEXT, human_name='b2')[0]
    assert not os.path.exists(b.data['file'][0]), 'Non Managed S3 file w/ remote should be copied to remote'
    b.pull(localize=True)
    assert os.path.exists(b.data['file'][0]), 'Non Managed S3 file after pull should be copied to local'

    # Get objects from remote
    objects = s3_client.list_objects(Bucket=TEST_BUCKET_OTHER)
    keys = [o['Key'] for o in objects['Contents']]
    keys = [key.split('/')[-1] for key in keys]

    # Make sure files exist in S3
    for output_file in ['test.parquet']:
        assert output_file in keys, 'Pipeline should have pushed file'


@moto.mock_aws
def test_remote_no_push_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Bind remote context
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    with pytest.raises(Exception) as e:
        dlapi.apply(TEST_CONTEXT, ManagedS3)


@moto.mock_aws
def test_remote_no_push_non_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)
    s3_resource.create_bucket(Bucket=TEST_BUCKET_OTHER)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'
    objects = s3_client.list_objects(Bucket=TEST_BUCKET_OTHER)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Bind remote context
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    # Apply
    dlapi.apply(TEST_CONTEXT, NonManagedS3)
    print(api.cat(TEST_CONTEXT, 'b2'))

    # Local context should not contain file if a remote exists.
    b = api.search(TEST_CONTEXT, human_name='b2')[0]
    assert not os.path.exists(b.data['file'][0]), 'Non Managed S3 file w/ remote should be copied to remote'
    assert b.data['file'][0].startswith("s3://")


def test_no_remote_push_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    with pytest.raises(Exception) as e:
        dlapi.apply(TEST_CONTEXT, ManagedS3, incremental_push=True)


@moto.mock_aws
def test_no_remote_push_non_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)
    s3_resource.create_bucket(Bucket=TEST_BUCKET_OTHER)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'
    objects = s3_client.list_objects(Bucket=TEST_BUCKET_OTHER)
    assert 'Contents' not in objects, 'Bucket should be empty'

    dlapi.apply(TEST_CONTEXT, NonManagedS3, incremental_push=True)
    print(api.cat(TEST_CONTEXT, 'b2'))
    assert len(api.search(TEST_CONTEXT)) == 1, 'One bundle should be present'

    assert os.path.exists(api.search(TEST_CONTEXT, human_name='b2')[0].data['file'][0]), \
        'Non Managed S3 file should be copied to local'


def test_no_remote_no_push_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    with pytest.raises(Exception) as e:
        dlapi.apply(TEST_CONTEXT, ManagedS3)


@moto.mock_aws
def test_no_remote_no_push_non_managed_s3():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)
    s3_resource.create_bucket(Bucket=TEST_BUCKET_OTHER)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'
    objects = s3_client.list_objects(Bucket=TEST_BUCKET_OTHER)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Apply
    dlapi.apply(TEST_CONTEXT, NonManagedS3)
    print(api.cat(TEST_CONTEXT, 'b2'))
    assert len(api.search(TEST_CONTEXT)) == 1, 'One bundle should be present'

    assert os.path.exists(api.search(TEST_CONTEXT, human_name='b2')[0].data['file'][0]), \
        'Non Managed S3 file should be copied to local'


if __name__ == '__main__':
    pytest.main([__file__])












#!/usr/bin/env python

# This Python script must be DEAD SIMPLE - it can *only* use standard
# Python modules. We relax this rule if you want to install from a setup.py
# file.

import argparse
import os
import re
import subprocess
import sys
import tempfile

from find_packages_from_setup import find_packages

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Select conda-provided Python packages from a pip-style requirements file'
    )
    parser.add_argument(
        'requirements',
        type=str,
        help='The requirements file',
    )
    parser.add_argument(
        '--no-non-conda',
        action='store_false',
        help='Do not output non-conda-provided packages (default is to output non-conda-provided packages on stderr)',
        dest='non_conda',
    )
    args = parser.parse_args()

    if not os.path.exists(args.requirements):
        raise RuntimeError('Failed to find file {}'.format(args.requirements))

    with tempfile.NamedTemporaryFile() as sanitized_file:
        # Create a sanitized requirements file that does not contain https
        # or -e pseudo-package references. Because Anaconda hates pip.
        packages = []
        if os.path.basename(args.requirements) == 'setup.py':
            packages = find_packages(args.requirements)
        else:
            with open(args.requirements, 'r') as requirements_file:
                packages = [p.rstrip() for p in requirements_file if not (p.startswith('git+') or p.startswith('http') or p.startswith('-e'))]
        sanitized_file.write('\n'.join(packages))
        sanitized_file.seek(0)
        # Use the conda installer to figure out which packages Continuum does
        # and does not host.
        non_conda_packages_p = subprocess.Popen(
            ['conda', 'install', '--dry-run', '--file', sanitized_file.name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True
        )
        # The installer lists the non-conda packages in stderr in the following
        # format:
        #   - package_name [version constraint operators (<>=)] version_num
        #
        # Oh yeah, and Python package names are case-insensitive.
        non_conda_packages_raw = non_conda_packages_p.communicate()[1]
        non_conda_packages_p.wait()
        non_conda_packages = []
        if non_conda_packages_p.returncode != 0:
            non_conda_packages = [p.lstrip(' -').split()[0].lower() for p in non_conda_packages_raw.split('\n') if re.match(r'^\s*-', p)]
        # Drop the version constraint, and filter out packages not provided by
        # Continuum.
        sanitized_file.seek(0)
        if len(non_conda_packages) > 0:
            non_conda_packages_re = r'^({})[<>=]*'.format('|'.join(non_conda_packages))
            conda_packages = [p.rstrip() for p in sanitized_file if not re.match(non_conda_packages_re, p.lower())]
        else:
            conda_packages = [p.rstrip() for p in sanitized_file]
        # Print the filtered requirements in a format usable with the conda
        # installer.
        print '\n'.join(conda_packages)
        if args.non_conda:
            print >> sys.stderr, '\n'.join(non_conda_packages)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import luigi
from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT

""" This set of tests verifies the behavior of Disdat's re-execution / data re-use logic.
 
 Luigi has a simple re-run logic: do not re-run tasks when the targets returned by the output function exist.
 
 Note that because task parameters should define up-stream dependencies, a given task should only re-use its result
 when its upstream required tasks have also re-used their results.  If a tasks parameters have not changed, its dependencies
 should not change, those cached results may be used, and so on.  
 
 But tasks with the same parameters may be re-executed for a variety of real-world reasons.  First, the task code may have 
 changed.  Or the task may be reading from an external data source that has changed (a database table).  While this 
 second case should be handled by the programmer (parameterizing by the time of db access), the programmer may forget or 
 the time unit may not be of sufficient granularity.   E.g., parametrize by day read, but the database changes more frequently.
 Or an external up-stream input, produced by another pipeline, has been updated.  
 
 In addition, it is possible for a layer above Disdat (or Luigi), could dynamically create tasks, and parameterize them
 outside of the normal pattern of "right-to-left" or "last task passes parameters to first task" pattern. 
 
 Finally, Disdat de-couples absolute file paths (used in Luigi) from the logical name of a task's output.  What this 
 means is that absolute file paths alone do not dictate whether a task should re-use that output.  Instead, whether any 
 *version* of a parameterized task's output exists determines if it should re-run.  
 
 In all of these cases, the downstream task should be re-run when new versions of upstream outputs exist.  Disdat builds 
 upon Luigi's straightforward logic to determine whether to re-execute or re-use existing outputs in these conditions. 
 
 In particular, Disdat uses the processing_name as the notion of versioning "sameness" for re-use.  The processing name
 is, a summary of the current task's parameters and a summary of the parameters of its upstream tasks.  In essence, it is 
 is analogous to a Merkle Tree, summarizing the state of the world used to run a single task.   
 
 Note that UUIDs are "lineage" names.  They tell us exactly the data used to produce an output.      
  
  Tests: * means new params
  1.) Force and Force all are handled by test_force_one_and_all.py
  2.) Run A, Run A, should re-use
  3.) Run A, Run A*, should re-run
  4.) Run A->B, Re-run A*.  Run A->B, nothing should run. 
  5.) Run A->B, re-run A (force), Run A->B, B should re-run. 
  6.) Run A->B, Re-run A*.  Run A*->B, B should re-run. 
  7.) Run A->B->C,  Run A*->B.   Run A->B->C, nothing should run
  8.) Run A->B->C,  Run A*->B.   Run A*->B->C, C should re-run
  9.) Run A,B -> C(a=A,b=B) Run A,B -> C(a=B,b=A), C should run both times.  
   
 """

WORKERS = 2


def test_A2_A3(run_test):
    """
    2.) Run A, Run A, should re-use
    3.) Run A, Run A*, should re-run
    """

    result = dlapi.apply(TEST_CONTEXT, A, workers=WORKERS)
    assert result['did_work'] is True
    first_A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    result = dlapi.apply(TEST_CONTEXT, A, workers=WORKERS)
    assert result['did_work'] is False
    second_A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    assert first_A_uuid == second_A_uuid
    assert len(api.search(TEST_CONTEXT, 'A')) == 1

    # Mod args, should re-run
    result = dlapi.apply(TEST_CONTEXT, A, params={'a': 2,'b': 3}, workers=WORKERS)
    assert result['did_work'] is True
    next_A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    assert next_A_uuid != second_A_uuid
    assert len(api.search(TEST_CONTEXT, 'A')) == 2


def test_AB4(run_test):
    """
    4.) Run A->B, Re-run A*.  Run A->B, nothing should run.
    """

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True

    result = dlapi.apply(TEST_CONTEXT, A, params={'a':2, 'b':3}, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is False


def test_AB5(run_test):
    """
    5.) Run A->B, re-run A (force), Run A->B, B should re-run.
    """
    api.delete_context(TEST_CONTEXT)
    api.context(TEST_CONTEXT)

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True

    result = dlapi.apply(TEST_CONTEXT, A, force=True, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True


def test_AB6(run_test):
    """
    6.) Run A->B, Re-run A*.  Run A*->B, B should re-run.

    Args:
        run_test:

    Returns:

    """

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    B_uuid = api.get(TEST_CONTEXT, 'B').uuid

    result = dlapi.apply(TEST_CONTEXT, APrime, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    APrime_uuid = api.get(TEST_CONTEXT, 'APrime').uuid

    def custom_B_requires(self):
        self.add_dependency('a', APrime, params={})

    old_requires = B.pipe_requires
    B.pipe_requires = custom_B_requires

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    assert APrime_uuid == api.get(TEST_CONTEXT, 'APrime').uuid
    assert B_uuid != api.get(TEST_CONTEXT, 'B').uuid

    B.pipe_requires = old_requires


def test_ABC7(run_test):
    """
    7.) Run A->B->C,  Run A*->B.   Run A->B->C, nothing should run

    Args:
        run_test:

    Returns:

    """

    result = dlapi.apply(TEST_CONTEXT, C, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    B_uuid = api.get(TEST_CONTEXT, 'B').uuid

    def custom_B_requires(self):
        self.add_dependency('a', APrime, params={})

    old_requires = B.pipe_requires
    B.pipe_requires = custom_B_requires

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    assert B_uuid != api.get(TEST_CONTEXT, 'B').uuid  # should have a new B

    B.pipe_requires = old_requires

    result = dlapi.apply(TEST_CONTEXT, C, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is False


def test_ABC8(run_test):
    """
    8.) Run A->B->C,  Run A*->B.   Run A*->B->C, C should re-run

    Args:
        run_test:

    Returns:

    """

    result = dlapi.apply(TEST_CONTEXT, C, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    B_uuid = api.get(TEST_CONTEXT, 'B').uuid

    def custom_B_requires(self):
        self.add_dependency('a', APrime, params={})

    old_requires = B.pipe_requires
    B.pipe_requires = custom_B_requires

    result = dlapi.apply(TEST_CONTEXT, B, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    assert B_uuid != api.get(TEST_CONTEXT, 'B').uuid  # should have a new B
    B_uuid = api.get(TEST_CONTEXT, 'B').uuid
    APrime_uuid = api.get(TEST_CONTEXT, 'APrime').uuid

    result = dlapi.apply(TEST_CONTEXT, C, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    assert B_uuid == api.get(TEST_CONTEXT, 'B').uuid
    assert APrime_uuid == api.get(TEST_CONTEXT, 'APrime').uuid

    B.pipe_requires = old_requires


def test_bundle_depsABC9(run_test):
    """
    10.) Run A,B -> C(a=A,b=B) Run A,B -> C(a=B,b=A), C should run both times.

    Args:
        run_test:

    Returns:

    """

    def custom_C_requires(self):
        self.add_dependency('a', A, params={})
        self.add_dependency('b', B, params={})

    old_requires = C.pipe_requires
    C.pipe_requires = custom_C_requires

    result = dlapi.apply(TEST_CONTEXT, C, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    B_uuid = api.get(TEST_CONTEXT, 'B').uuid
    C_uuid = api.get(TEST_CONTEXT, 'C').uuid

    def custom_C_requires_swap(self):
        self.add_dependency('a', B, params={})
        self.add_dependency('b', A, params={})

    C.pipe_requires = custom_C_requires_swap

    result = dlapi.apply(TEST_CONTEXT, C, workers=WORKERS)
    assert result['success'] is True
    assert result['did_work'] is True
    assert A_uuid == api.get(TEST_CONTEXT, 'A').uuid
    assert B_uuid == api.get(TEST_CONTEXT, 'B').uuid
    assert C_uuid != api.get(TEST_CONTEXT, 'C').uuid

    C.pipe_requires = old_requires


class A(PipeTask):

    a = luigi.IntParameter(default=1)
    b = luigi.IntParameter(default=2)

    def pipe_requires(self):
        pass

    def pipe_run(self):
        return self.a+self.b


class APrime(PipeTask):

    a = luigi.IntParameter(default=2)
    b = luigi.IntParameter(default=1)

    def pipe_requires(self):
        pass

    def pipe_run(self):
        return self.a+self.b


class B(PipeTask):

    a = luigi.IntParameter(default=3)
    b = luigi.IntParameter(default=4)

    def pipe_requires(self):
        self.add_dependency('a', A, params={})

    def pipe_run(self, a):
        return a + self.a + self.b


class C(PipeTask):

    a = luigi.IntParameter(default=5)
    b = luigi.IntParameter(default=6)

    def pipe_requires(self):
        self.add_dependency('b', B, params={})

    def pipe_run(self, **kwargs):
        input_sum = sum([v for v in kwargs.values()])
        return input_sum + self.b + self.a


if __name__ == '__main__':
    pytest.main([__file__])





#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from disdatluigi.pipe import PipeTask
import luigi

""" These are simple tasks used for test_api_run """

COMMON_DEFAULT_ARGS=[10, 100, 1000]


class B(PipeTask):
    """ B required by A """
    int_array = luigi.ListParameter(default=None)

    def pipe_run(self):
        print ("B saving type [{}]".format(type(self.int_array)))
        return self.int_array


class A(PipeTask):
    """ A is the root task"""
    int_array = luigi.ListParameter(default=COMMON_DEFAULT_ARGS)

    def pipe_requires(self):
        self.add_dependency('b', B, {'int_array': self.int_array})

    def pipe_run(self, b=None):
        print ("Saving the sum of B {}".format(b))
        print ("A got type [{}]".format(type(b)))
        return sum(list(b))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Disdat comes with a CLI.  It looks for the presence of certain packages that might extend that CLI.
One of them is "disdatluigi".  It then looks for a module "cli_extension.py", and the "add_arg_parser" method.
The 3rd-party package is then able to add commands to the CLI so users have a single interface to using
Disdat on the command line.

Disdat-Luigi needs to add a number of subparsers.  These include:
apply.py:
dockerize.py:
run.py:
"""

import disdatluigi.apply
import disdatluigi.dockerize
import disdatluigi.run
from disdatluigi import logger as _logger


def add_arg_parser(subparsers):
    _logger.debug("Disdat-Luigi: extending Disdat CLI...")
    ls_p = subparsers.add_parser('luigi')
    ls_p.set_defaults(func=lambda args: DisdatConfig.init())

    disdatluigi.apply.add_arg_parser(subparsers)
    disdatluigi.dockerize.add_arg_parser(subparsers)
    disdatluigi.run.add_arg_parser(subparsers)
    return



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import luigi

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT


""" Test ability to call pipelines in pipelines 
Test with workers=1 and workers>1
"""


class A(PipeTask):
    n = luigi.IntParameter()

    def pipe_requires(self):
        self.set_bundle_name("A[{}]".format(self.n))
        self.add_dependency("shark", B, params={'n': self.n+1})

    def pipe_run(self, shark=None):
        return self.n * 2


class B(A):
    def pipe_requires(self):
        self.set_bundle_name("B[{}]".format(self.n))


class Root(PipeTask):
    """
    Call another pipeline internally.
    """
    n = luigi.IntParameter(default=0)
    workers = luigi.IntParameter(default=1)
    inception = luigi.BoolParameter(default=False)

    def pipe_requires(self):
        self.set_bundle_name('root')
        self.add_dependency('a', A, params={'n': self.n+1})
        self.add_dependency('b', B, params={'n': self.n+1})

    def pipe_run(self, a=None, b=None):
        if self.inception:
            dlapi.apply(TEST_CONTEXT, Root, output_bundle="inception_result", params={'n': 100}, workers=self.workers)
            b = api.get(TEST_CONTEXT, "inception_result" )
            return b.data
        else:
            return a + b


def _inception_pipeline(inner_workers, outer_workers):
    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'
    dlapi.apply(TEST_CONTEXT, Root, output_bundle="test_root", params={'inception': True,
                                                                     'workers': inner_workers},
              workers=outer_workers)

    b = api.get(TEST_CONTEXT, "test_root")

    print("FINAL DATA {}".format(b.data))

def test_inception_1_1():
    _inception_pipeline(1, 1)

def test_inception_3_1():
    _inception_pipeline(3, 1)

def test_inception_1_3():
    _inception_pipeline(1, 3)

def test_inception_3_3():
    _inception_pipeline(3, 3)


if __name__ == '__main__':
    api.delete_context(TEST_CONTEXT)
    api.context(TEST_CONTEXT)
    test_inception_3_3()
    #api.apply(TEST_CONTEXT, A, params={'n': 1})


#!/usr/bin/env python

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
AWS SageMaker entrypoint wrapper for Disdatified pipelines.
"""

import argparse
import json
import logging
import os
import sys
import docker_ep

from multiprocessing import Process

_HELP = """ AWS SageMaker Disdat pipeline wrapper. This script will call the main entrypoint
to execute the pipeline, parsing arguments from hyperparameter.json.
"""

_logger = logging.getLogger(__name__)

# This is a hard-coded path that will be present in the container according to AWS SageMaker
# https://docs.aws.amazon.com/sagemaker/latest/dg/your-algorithms-training-algo.html
_HYPERPARAMETERS = "/opt/ml/input/config/hyperparameters.json"


def add_argument_help_string(help_string, default=None):
    if default is None:
        return '{}'.format(help_string)
    else:
        return "{} (default '{}')".format(help_string, default)


def train():

    with open(_HYPERPARAMETERS) as hp:
        args = json.load(hp)

    arglist = json.loads(args['arglist'])

    _logger.info("Disdat SageMaker Train calling entrypoint with json loads arglist {}".format(arglist))

    p = Process(target=docker_ep.main, args=[arglist,])
    p.start()
    p.join()
    return p.exitcode == 0


if __name__ == '__main__':
    """ SageMaker invokes the container with 'train' or 'serve'.
    Train jobs support arbitrary 'hyperparameter' params inside a json blob.
    We read the json and interpret them as arguments to the Disdat entrypoint.

    Note:
    1.) We ignore the input S3 path (inputs come as bundles)
    2.) We use the output S3 path to store the output bundle context, remote, name and UUID
    """

    parser = argparse.ArgumentParser(
        description=_HELP,
    )

    parser.add_argument(
        'purpose',
        type=str,
        help="'train' or 'serve'",
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    _logger.setLevel(logging.INFO)

    if args.purpose == 'train':
        if not train():
            _logger.error("Disdat SageMaker train entrypoint failed.")
            sys.exit(os.EX_IOERR)
    elif args.purpose == 'serve':
        _logger.warning("Disdat does not yet support SageMaker serve.")
        sys.exit(os.EX_UNAVAILABLE)
    else:
        _logger.error("Disdat SageMaker invoked entrypoint with {}, not 'train' or 'serve'".format(args.purpose))
        sys.exit(os.EX_USAGE)

    sys.exit(os.EX_OK)




#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import pytest
import luigi
import disdat.api as dsdt_api
from disdatluigi.pipe import PipeTask
import disdatluigi.api as api
from tests.conftest import TEST_CONTEXT

test_luigi_args_data = {'str_arg': 'some string',
                    'int_arg': 10,
                    'list_arg': [1,3,5],
                    'list_str_arg': ['farg','barg'],
                    'dict_float_arg': {'farg': 0.01, 'barg': 3.14},
                    'date_arg': datetime.date(2020,4,1)}


class ArgTask(PipeTask):
    str_arg = luigi.Parameter(default=None)
    int_arg = luigi.IntParameter(default=None)
    list_arg = luigi.ListParameter(default=None)
    list_str_arg = luigi.ListParameter(default=None)
    dict_float_arg = luigi.DictParameter(default=None)
    date_arg = luigi.DateParameter(default=None)

    def pipe_run(self):
        return True


def test_luigi_args():
    """ Create a task, store args, retrieve from bundle api.
    Pass in python objects as the values for Luigi parameters.
    Stored as serialized json objects.   Bundle presents the parameters
    as the serialized objects (Disdat isn't aware they were Luigi serialized).
    """

    api.apply(TEST_CONTEXT, ArgTask, output_bundle='output', params=test_luigi_args_data)
    b = dsdt_api.get(TEST_CONTEXT, 'output')
    found_p = {}
    for k, p in b.params.items():
        attribute = getattr(ArgTask, k)
        found_p[k] = attribute.parse(p)
    assert(found_p == test_luigi_args_data)


if __name__ == "__main__":
    pytest.main([__file__])



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Resource
========
Package resource manager with the same basic API as pkg_resources.

Advantages:
* Loads much faster. pkg_resources takes 200ms to load, this takes 0ms to load.
* Allows you to pass in that actual loaded module instead of using a string.
  This allows for better refactoring.

"""
import pkgutil
import sys
import os
import types


def filename(package, resource):

    if isinstance(package, types.ModuleType):
        mod = package
    else:
        loader = pkgutil.get_loader(package)
        if loader is None or not hasattr(loader, 'get_data'):
            return None
        mod = sys.modules.get(package) or loader.load_module(package)
        if mod is None or not hasattr(mod, '__file__'):
            return None

    parts = resource.split('/')
    parts.insert(0, os.path.dirname(mod.__file__))
    return os.path.join(*parts)


def exists(package, resource):
    return os.path.exists(filename(package, resource))


def stream(package, resource):
    return open(filename(package, resource), 'rb')


def isdir(package, resource):
    os.path.isdir(filename(package, resource))


def listdir(package, resource):
    os.listdir(filename(package, resource))


def string(package, resource):
    with open(filename(package, resource), 'rb') as handle:
        return handle.read()



'''
Created on Sep 12, 2017

@author: twong
'''

import os


def which(cmd_name):
    '''Get the full path to an external command executable.

    :param cmd_name: The command name
    :return: The full path to the command executable, or `None` if the
    executable is not on the O/S path
    :rtype: str
    '''
    paths = os.environ['PATH'].split(os.pathsep)
    for p in paths:
        cmd_fq_name = os.path.join(p, cmd_name)
        if os.path.exists(cmd_fq_name) and os.access(cmd_fq_name, os.X_OK):
            return cmd_fq_name
    return None



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import luigi
import pytest

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from disdatluigi.common import ApplyError
from tests.conftest import TEST_CONTEXT

EXT_BUNDLE_NAME='ext_bundle_human_name'
BUNDLE_CONTENTS=list(range(9))
EXT_TASK_PARAM_VAL='this is a test value'


class ExternalPipeline(PipeTask):
    test_param = luigi.Parameter()

    """ External Pipeline """
    def pipe_requires(self):
        self.set_bundle_name('external_pipeline')

    def pipe_run(self):
        print ("ExternalPipeline called with parameter [{}]".format(self.test_param))
        return BUNDLE_CONTENTS


class PipelineA(PipeTask):
    test_param = luigi.Parameter(default=EXT_TASK_PARAM_VAL)
    throw_assert = luigi.BoolParameter(default=True)

    def pipe_requires(self):
        self.set_bundle_name('pipeline_a')
        b = self.add_external_dependency('ext_input', ExternalPipeline, {'test_param': self.test_param})
        if self.throw_assert:
            assert b is not None
            assert list(b.data) == BUNDLE_CONTENTS

    def pipe_run(self, ext_input=None):
        assert list(ext_input) == BUNDLE_CONTENTS
        return True


class PipelineB(PipeTask):
    ext_uuid = luigi.Parameter()

    def pipe_requires(self):
        self.set_bundle_name('pipeline_b')
        b = self.add_external_dependency('ext_input',
                                         ExternalPipeline,
                                         {},
                                         uuid=self.ext_uuid)
        assert b is not None
        assert list(b.data) == BUNDLE_CONTENTS

    def pipe_run(self, ext_input=None):
        assert list(ext_input) == BUNDLE_CONTENTS
        return True


class PipelineC(PipeTask):
    ext_name = luigi.Parameter()

    def pipe_requires(self):
        self.set_bundle_name('pipeline_b')
        b = self.add_external_dependency('ext_input',
                                         ExternalPipeline,
                                         {},
                                         human_name=self.ext_name)
        assert b is not None
        assert list(b.data) == BUNDLE_CONTENTS

    def pipe_run(self, ext_input=None):
        assert list(ext_input) == BUNDLE_CONTENTS
        return True


def create_bundle_from_pipeline():
    """ Run the internal pipeline, create a bundle, return the uuid
    """

    dlapi.apply(TEST_CONTEXT,
              ExternalPipeline,
              params={'test_param': EXT_TASK_PARAM_VAL},
              output_bundle=EXT_BUNDLE_NAME)
    b = api.get(TEST_CONTEXT, EXT_BUNDLE_NAME)
    return b.uuid


def test_ord_external_dependency(run_test):

    uuid = create_bundle_from_pipeline()

    dlapi.apply(TEST_CONTEXT, PipelineA)

    result = dlapi.apply(TEST_CONTEXT, PipelineA)
    assert result['success'] is True
    assert result['did_work'] is False


def test_uuid_external_dependency(run_test):

    uuid = create_bundle_from_pipeline()

    dlapi.apply(TEST_CONTEXT, PipelineB, params={'ext_uuid': uuid})

    result = dlapi.apply(TEST_CONTEXT, PipelineB, params={'ext_uuid': uuid})
    assert result['success'] is True
    assert result['did_work'] is False


def test_name_external_dependency(run_test):

    uuid = create_bundle_from_pipeline()

    dlapi.apply(TEST_CONTEXT, PipelineC, params={'ext_name': EXT_BUNDLE_NAME})

    result = dlapi.apply(TEST_CONTEXT, PipelineC, params={'ext_name': EXT_BUNDLE_NAME})
    assert result['success'] is True
    assert result['did_work'] is False


def test_ord_external_dependency_fail(run_test):
    """ Test ability to handle a failed lookup.
    Note: Disdat/Luigi swallows exceptions in tasks.  Here our tasks
    assert that they get back a bundle on their lookup.  If we catch it, then the
    test succeeds.

    Args:
        run_test:

    Returns:

    """

    uuid = create_bundle_from_pipeline()

    try:
        result = dlapi.apply(TEST_CONTEXT, PipelineA, params={'test_param': 'never run before'})
    except ApplyError as ae:
        print("ERROR: {}".format(ae))
        return


def test_uuid_external_dependency_fail(run_test):
    """ Test ability to handle a failed lookup.
    Note: Disdat/Luigi swallows exceptions in tasks.  Here our tasks
    assert that they get back a bundle on their lookup.  If we catch it, then the
    test succeeds.

    Args:
        run_test:

    Returns:

    """

    uuid = create_bundle_from_pipeline()
    try:
        result = dlapi.apply(TEST_CONTEXT, PipelineB, params={'ext_uuid': 'not a valid uuid'})
    except ApplyError as ae:
        print("ERROR: {}".format(ae))
        return


def test_name_external_dependency_fail(run_test):
    """ Test ability to handle a failed lookup.
    Note: Disdat/Luigi swallows exceptions in tasks.  Here our tasks
    assert that they get back a bundle on their lookup.  If we catch it, then the
    test succeeds.

    Args:
        run_test:

    Returns:

    """

    uuid = create_bundle_from_pipeline()
    try:
        result = dlapi.apply(TEST_CONTEXT, PipelineC, params={'ext_name': 'not a bundle name'})
    except ApplyError as ae:
        print("ERROR: {}".format(ae))
        return


if __name__ == '__main__':
    pytest.main([__file__]) # +"::test_ord_external_dependency_fail"])



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect
import logging
import os
import tempfile
import shutil

import docker
import disdat.utility.aws_s3 as aws
from disdat.fs import determine_pipe_version

import disdatluigi.resources
from disdatluigi.infrastructure import dockerizer
from disdatluigi.common import do_subprocess, make_project_image_name, make_project_repository_name
from disdatluigi.common import DisdatLuigiConfig, setup_exists, make_sagemaker_project_repository_name, make_sagemaker_project_image_name


_MODULE_NAME = inspect.getmodulename(__file__)

_DOCKERIZER_ROOT = os.path.dirname(inspect.getsourcefile(dockerizer))

_logger = logging.getLogger(__name__)


def _copy_in_dot_file(disdat_config, docker_context, dot_file_name, option_name, cli):
    """
    Copy in a dot file to a file with name "dot_file_name" into the docker context.
    The user might have put the src path in the disdat config file under the "option_name"

    Args:
        disdat_config:  A disdat config object
        docker_context:  The place we are storing the docker context for build
        dot_file_name:  The name of the file in the docker context
        option_name: the name of the option in disdatluigi.cfg
        cli (bool): whether we are called from cli

    Returns:
        None or if subprocess has output

    """
    retval = 0

    dot_file_path = os.path.join(docker_context, dot_file_name)
    if disdat_config.parser.has_option(_MODULE_NAME, option_name):
        dot_file = os.path.expanduser(disdat_config.parser.get(_MODULE_NAME, option_name))
        shutil.copy(dot_file, dot_file_path)
        print("Copying dot file {} into {}".format(dot_file, docker_context))
    else:
        touch_command = [
            'touch',
            dot_file_path
        ]
        retval = do_subprocess(touch_command, cli)

    return retval


def latest_container_id(pipeline_root, cli):
    """ Return the unique container image hash from the latest container made
    from this setup.py

    Args:
        pipeline_root (str): The path to the setup.py file that defines the container
        cli (bool): if called from cli

    Returns:
        (str): The full docker container image hash

    """
    setup_file = os.path.join(pipeline_root, 'setup.py')
    pipeline_image_name = make_project_image_name(setup_file)
    docker_client = docker.from_env()

    try:
        img_obj = docker_client.images.get(pipeline_image_name)
    except (docker.errors.APIError, docker.errors.ImageNotFound) as er:
        if cli:
            print("Disdat unable to find image with project name {}".format(pipeline_image_name))
            raise # exit with code > 0
        return None

    id = img_obj.id.replace('sha256:','')

    if cli:
        print("{}".format(id))

    return id


def dockerize(pipeline_root,
              config_dir=None,
              os_type=None,
              os_version=None,
              build=True,
              push=False,
              sagemaker=False,
              cli=False
              ):
    """ Create a Docker image for running a pipeline.

    Args:
        pipeline_root: Root of the Python source tree containing the
        setuptools-style setup.py file.
        config_dir (str):  Configuration of image (.deb, requires.txt, etc.)
        os_type (str): OS type string
        os_version (str): Version of OS
        build (bool): Build the image (default True)
        push (bool): Push to registry listed in Disdat config file
        sagemaker (bool): Build a container for 'train' or 'serve' in SageMaker
        cli (bool): Whether dockerize was called from the CLI (True) or an API (False -- default)

    Returns:
        (int): 0 equals success, >0 for error

    """

    disdat_config = DisdatLuigiConfig.instance()

    # Get configuration parameters
    image_os_type = os_type if os_type is not None else disdat_config.parser.get(_MODULE_NAME, 'os_type')
    image_os_version = os_version if os_version is not None else disdat_config.parser.get(_MODULE_NAME, 'os_version')
    docker_context = None

    if docker_context is None:
        docker_context = tempfile.mkdtemp(suffix=_MODULE_NAME)
    docker_makefile = os.path.join(_DOCKERIZER_ROOT, 'Makefile')
    _logger.debug('Using Docker context {}'.format(docker_context))

    # Populate the Docker context with the template containing dockerfiles,
    # entrypoints, etc.
    rsync_command = [
        'rsync',
        '-aq',  # Archive mode, no non-error messages
        '--exclude', '*.pyc',  # Don't copy any compiled Python files
        os.path.join(_DOCKERIZER_ROOT, 'context.template', ''),
        docker_context,
    ]

    retval = do_subprocess(rsync_command, cli)
    if retval: return retval

    # PIP: Overwrite pip.conf in the context.template in your repo if they have set the option,
    # else just create empty file.
    # At this time, the Dockerfile always sets the PIP_CONFIG_FILE ENV var to this file.
    retval = _copy_in_dot_file(disdat_config, docker_context, "pip.conf", "dot_pip_file", cli)
    if retval: return retval

    # ODBC: Overwrite the .odbc.ini in the context.template in your repo if the user set the option,
    # else just create empty file.
    # At this time, the Dockerfile always sets the ODBCINI var to this file.
    retval = _copy_in_dot_file(disdat_config, docker_context, "odbc.ini", "dot_odbc_ini_file", cli)
    if retval: return retval

    setup_file = os.path.join(pipeline_root, 'setup.py')

    if not setup_exists(setup_file):
        return 1

    pipeline_image_name = make_project_image_name(setup_file)

    DEFAULT_DISDAT_HOME = os.path.join('/', *os.path.dirname(disdatluigi.dockerize.__file__).split('/')[:-1])
    DISDAT_HOME = os.getenv('DISDAT_HOME', DEFAULT_DISDAT_HOME)

    if build:
        pipe_version = determine_pipe_version(pipeline_root)
        build_command = [
            'make',  # XXX really need to check that this is GNU make
            '-f', docker_makefile,
            'PIPELINE_IMAGE_NAME={}'.format(pipeline_image_name),
            'DISDAT_DOCKER_CONTEXT={}'.format(docker_context),
            'DISDAT_ROOT={}'.format(os.path.join(DISDAT_HOME)),  # XXX YUCK
            'PIPELINE_ROOT={}'.format(pipeline_root),
            'OS_TYPE={}'.format(image_os_type),
            'OS_VERSION={}'.format(image_os_version),
            'GIT_HASH={}'.format(pipe_version.hash),
            'GIT_BRANCH={}'.format(pipe_version.branch),
            'GIT_FETCH_URL={}'.format(pipe_version.url),
            'GIT_TIMESTAMP={}'.format(pipe_version.tstamp),
            'GIT_DIRTY={}'.format(pipe_version.dirty),
        ]

        _logger.debug("pipeline root = {} build command = {}".format(pipeline_root, build_command))

        if config_dir is not None:
            build_command.append('CONFIG_ROOT={}'.format(config_dir))
        if sagemaker:
            build_command.append('SAGEMAKER_TRAIN_IMAGE_NAME={}'.format(make_sagemaker_project_image_name(setup_file)))
            build_command.append('sagemaker')
        retval = do_subprocess(build_command, cli)
        if retval: return retval

    if push:
        docker_client = docker.from_env()

        repository_name_prefix = None

        if disdat_config.parser.has_option('docker', 'repository_prefix'):
            repository_name_prefix = disdat_config.parser.get('docker', 'repository_prefix')
        if sagemaker:
            repository_name = make_sagemaker_project_repository_name(repository_name_prefix, setup_file)
            pipeline_image_name = make_sagemaker_project_image_name(setup_file)
        else:
            repository_name = make_project_repository_name(repository_name_prefix, setup_file)

        # Figure out the fully-qualified repository name, i.e., the name
        # including the registry.
        if disdat_config.parser.has_option('docker', 'registry'):
            registry_name = disdat_config.parser.get('docker', 'registry').strip('/')
            if registry_name == '*ECR*':
                policy_resource_name = None
                if disdat_config.parser.has_option('docker', 'ecr_policy'):
                    policy_resource_name = disdat_config.parser.get('docker', 'ecr_policy')
                fq_repository_name = aws.ecr_create_fq_respository_name(
                    repository_name,
                    policy_resource_package=disdatluigi.resources,
                    policy_resource_name=policy_resource_name
                )
            else:
                fq_repository_name = '{}/{}'.format(registry_name, repository_name)
        else:
            if cli:
                raise RuntimeError("No registry present for push to succeed")
            else:
                return 1

        auth_config = None
        if disdat_config.parser.has_option('docker', 'ecr_login') or registry_name == '*ECR*':
            auth_config = aws.ecr_get_auth_config()
        docker_client.api.tag(pipeline_image_name, fq_repository_name)
        for line in docker_client.images.push(fq_repository_name, auth_config=auth_config, stream=True):
            if b'error' in line:
                if cli:
                    raise RuntimeError(line)
                else:
                    return 1
            else:
                if cli: print(line)

    return 0


def add_arg_parser(parsers):
    dockerize_p = parsers.add_parser('dockerize', description="Dockerizer a particular transform.")
    dockerize_p.add_argument(
        '--config-dir',
        type=str,
        default=None,
        help="A directory containing configuration files for the operating system within the Docker image",
    )
    dockerize_p.add_argument(
        '--os-type',
        type=str,
        default=None,
        help='The base operating system type for the Docker image',
    )
    dockerize_p.add_argument(
        '--os-version',
        type=str,
        default=None,
        help='The base operating system version for the Docker image',
    )
    dockerize_p.add_argument(
        '--push',
        action='store_true',
        help="Push the image to a remote Docker registry (default is to not push; must set 'docker_registry' in Disdat config)",
    )
    dockerize_p.add_argument(
        '--get-id',
        action='store_true',
        help="Do not build, only return latest container image ID",
    )
    dockerize_p.add_argument(
        '--sagemaker',
        action='store_true',
        default=False,
        help="Create a Docker image executable as a SageMaker container.",
    )
    dockerize_p.add_argument(
        '--no-build',
        action='store_false',
        help='Do not build an image (only copy files into the Docker build context)',
        dest='build',
    )
    dockerize_p.add_argument(
        "pipeline_root",
        type=str,
        help="Root of the Python source tree containing the user-defined transform; must have a setuptools-style setup.py file"
    )
    dockerize_p.set_defaults(func=lambda args: dockerize_entry(cli=True, **vars(args)))
    return parsers


def dockerize_entry(cli=False, **kwargs):
    """Run the dockerize command with parameters from the command line or the api.

    Parameters:
        cli (bool): Whether this was called from the command line or the api, default False
        **kwargs

    Returns:
        (int): 0 for success, 1 for failure
    """

    if 'get_id' in kwargs and kwargs['get_id']:
        return latest_container_id(kwargs['pipeline_root'], cli)
    else:
        return dockerize(kwargs['pipeline_root'],
                         config_dir=kwargs['config_dir'],
                         os_type=kwargs['os_type'],
                         os_version=kwargs['os_version'],
                         build=kwargs['build'],
                         push=kwargs['push'],
                         sagemaker=kwargs['sagemaker'],
                         cli=cli
                         )


if __name__ == "__main__":
    import api
    api.dockerize('/Users/kyocum/Code/anomaly-detection-service/anomaly', 'pipeline.pipeline.Train', push=True)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import luigi
import pandas as pd
import numpy as np
from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi

from tests.conftest import TEST_CONTEXT


def test(run_test):
    """ Purpose of this test is to have one task that produces a bundle.
    And another task that requires it.

    1.) Run DataMaker which runs PreMaker
    2.) Assert that those ran, and remove PreMaker
    3.) run Root which needs DataMaker (external dep) and PreMaker
    4.) assert that premaker re-ran and Root ran successfully (getting external dependency)

    """

    api.context(TEST_CONTEXT)

    dlapi.apply(TEST_CONTEXT, DataMaker, params={'int_array': [1000, 2000, 3000]})

    b = api.get(TEST_CONTEXT, 'PreMaker')
    assert(b is not None)
    pm_uuid = b.uuid
    b.rm()

    dlapi.apply(TEST_CONTEXT, Root)

    b = api.get(TEST_CONTEXT, 'PreMaker')
    assert(b is not None)
    assert(b.uuid != pm_uuid)

    b = api.get(TEST_CONTEXT, 'Root')
    assert(b is not None)

    api.delete_context(TEST_CONTEXT)


class DataMaker(PipeTask):
    """ Run this by itself.
    Then B requires DataMaker as external, and A. """

    int_array = luigi.ListParameter(default=[1, 2, 3, 5, 8])

    def pipe_requires(self):
        self.set_bundle_name("DataMaker")
        self.add_dependency('premaker', PreMaker, params={})
        return

    def pipe_run(self, premaker=None):

        return np.array(self.int_array)


class PreMaker(PipeTask):

    printme = luigi.Parameter(default="snarky")

    def pipe_requires(self):
        return

    def pipe_run(self):

        print("Task premaker says {}".format(self.printme))

        return pd.DataFrame({'fark': np.random.randint(100, size=10), 'bark': np.random.randint(10, size=10)})


class Root(PipeTask):

    def pipe_requires(self):
        self.add_dependency('premaker', PreMaker, params={})
        self.add_external_dependency('datamaker', DataMaker, {'int_array': [1000, 2000, 3000]})

    def pipe_run(self, premaker=None, datamaker=None):
        print ("Root received a datamaker {}".format(datamaker))
        return


if __name__ == '__main__':
    pytest.main([__file__])
    #test()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import luigi
from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT

TEST_NAME    = 'test_bundle'


def test(run_test):
    """ This tests if mark_force works for tasks.
    We have two tasks. One depends on the other.  The upstream is marked
    mark_force and should always run.
    """

    def run_and_get(name, do_ext=False):
        dlapi.apply(TEST_CONTEXT, A_2, params={'set_ext_dep': do_ext})
        b = api.get(TEST_CONTEXT, 'B')
        print ("Run {}: b.creation_date {} b.uuid {}".format(name, b.creation_date, b.uuid))
        return b

    b = run_and_get("One")
    first_uuid = b.uuid

    b = run_and_get("Two")
    assert(first_uuid != b.uuid)
    second_uuid = b.uuid

    b = run_and_get("Three", do_ext=True)
    assert(second_uuid == b.uuid)


class B_2(PipeTask):

    def pipe_requires(self):
        self.set_bundle_name("B")
        self.mark_force()
        return

    def pipe_run(self):
        print ("Task B finished.")

        return True


class A_2(PipeTask):
    set_ext_dep = luigi.BoolParameter(default=False)

    def pipe_requires(self):
        if self.set_ext_dep:
            self.add_external_dependency('B', B_2, params={})
        else:
            self.add_dependency('B', B_2, {})

    def pipe_run(self, B=None):
        print ("Task A finished.")
        return


if __name__ == '__main__':
    test()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import logging
import contextlib

import disdatluigi


# Disdat-Luigi global logger
logger = logging.getLogger(disdatluigi.__name__)
logger.addHandler(logging.NullHandler())

# External loggers
luigi_logger = logging.getLogger('luigi-interface')
boto3_logger = logging.getLogger('boto3')
botocore_logger = logging.getLogger('botocore')


@contextlib.contextmanager
def context(level=logging.INFO, stream=sys.stdout):
    """
    A context handler for logging debug information

    Args:
        level (int): The logging package log level to user within the context.
        stream (io.IO): The stream to write log messages to.

    Returns:
        logging.Logger: The logger to debug with
    """
    if level is None:
        yield logger
        return

    fmt = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler = logging.StreamHandler(stream)
    handler.setFormatter(fmt)
    prior = logger.getEffectiveLevel()
    logger.addHandler(handler)
    handler.setLevel(level)
    logger.setLevel(level)
    yield logger
    logger.removeHandler(handler)
    logger.setLevel(prior)


def enable(level=logging.INFO, stream=sys.stdout):
    """
    Sets the default package-wide configuration.

    Args:
        level (int): The logging package log level to user within the context.
        stream (io.IO): The stream to write log messages to.
    """
    handler = logging.StreamHandler(stream)
    logger.addHandler(handler)
    logger.setLevel(level)
